from abc import ABC, abstractmethod

from .collectors import Collector, MutCollector
from .selectors import ListSelector, Selector


class Registry[K, T](Collector[K, T], Selector[K, T], ABC): ...


class MutRegistry[K, T](Registry[K, T], MutCollector[K, T], ABC):
    @abstractmethod
    def clear(self) -> None: ...


class ListRegistry[K, T](Registry[K, T], ListSelector[K, T], ABC): ...


class MutListRegistry[K, T](MutRegistry[K, T], ListRegistry[K, T], ABC): ...



import enum
from datetime import datetime
from typing import override

from broken.controllers import Condition, ControllerState

from greyhorse.app.abc.functional.factories import MutFactoryRegistry
from greyhorse.app.abc.resources.common import ResourceEventKind
from greyhorse.app.abc.resources.operators import operator
from greyhorse.app.abc.resources.providers import ProviderEvent, provider_listener
from greyhorse.app.abc.services import ServiceState
from greyhorse.app.entities.controllers import SyncController
from greyhorse.app.entities.services import SyncService
from greyhorse.maybe import Maybe
from greyhorse.result import Ok, Result

from ..common.functional import FunctionalOperator
from ..common.resources import DictResContext, MutDictResContext


class TestLifetime(enum.IntEnum):
    DOMAIN = enum.auto(0)
    OPERATION = enum.auto()


class FunctionalOperatorImpl(FunctionalOperator):
    def __init__(self, ctx: DictResContext, mut_ctx: MutDictResContext) -> None:
        self.ctx = ctx
        self.mut_ctx = mut_ctx

    @override
    def add_number(self, value: int) -> Result[None, str]:
        with self.mut_ctx as ctx:
            ctx['number'] = value
            self.mut_ctx.apply()

        return Ok()

    @override
    def get_number(self) -> Result[int, str]:
        with self.ctx as ctx:
            value = Maybe(ctx.get('number'))

        return value.ok_or('Number is not initialized')

    @override
    def remove_number(self) -> Result[bool, str]:
        with self.mut_ctx as ctx:
            value = Maybe(ctx.pop('number', None))
            self.mut_ctx.apply()

        return value.ok_or('Number is not initialized').map(lambda _: True)


class DictResourceCtrl(SyncController):
    def __init__(self, factories: MutFactoryRegistry) -> None:
        super().__init__()
        self._factories = factories

    @override
    def setup(self) -> None:
        cond = Condition(
            name='DictResContext',
            status=self._factories.has_factory(DictResContext),
            permanent=True,
            transition_dt=datetime.now(),
        )
        self._set_condition(cond)
        cond = Condition(
            name='MutDictResContext',
            status=self._factories.has_factory(MutDictResContext),
            permanent=True,
            transition_dt=datetime.now(),
        )
        self._set_condition(cond)
        self._set_state(ControllerState.Succeeded)

    @override
    def teardown(self) -> None:
        self._reset_condition('DictResContext')
        self._reset_condition('MutDictResContext')
        self._set_state(ControllerState.Finalized)

    @provider_listener(DictResContext, fragment='frag1')
    def listener_1(self, event: ProviderEvent) -> None:
        if event.kind == ResourceEventKind.SETUP:
            f = event.context.__enter__()
            self._factories.add_factory(DictResContext, f)
        elif event.kind == ResourceEventKind.TEARDOWN:
            self._factories.remove_factory(DictResContext)
            event.context.__exit__()

    @provider_listener(MutDictResContext, fragment='frag1')
    def listener_2(self, event: ProviderEvent) -> None:
        if event.kind == ResourceEventKind.SETUP:
            f = event.context.__enter__()
            self._factories.add_factory(MutDictResContext, f)
        elif event.kind == ResourceEventKind.TEARDOWN:
            self._factories.remove_factory(MutDictResContext)
            event.context.__exit__()


class FunctionalService(SyncService):
    def __init__(self) -> None:
        super().__init__()

    @override
    def setup(self) -> None:
        self._set_state(ServiceState.Ready(started=False))

    @override
    def teardown(self) -> None:
        self._set_state(ServiceState.Unready(started=False, message='stopped'))

    # @operator(scope=TestLifetime.OPERATION, fragment='frag1')
    # @operator(fragment='frag1')

    op_1 = operator(
        for_type=FunctionalOperator, scope=TestLifetime.OPERATION, fragment='frag1'
    )(FunctionalOperatorImpl)



from collections.abc import Callable, Collection, Mapping
from dataclasses import dataclass, field
from typing import Any, override

from greyhorse.app.abc.functional.registries import MutFactoryRegistry
from greyhorse.factory import Factory
from greyhorse.maybe import Just, Maybe, Nothing

from ..collections.registries import MutDictRegistry


@dataclass(slots=True, frozen=True, kw_only=True)
class _CovariantFactoryFilter[T]:
    key: type[T]
    filters: Mapping[str, Any] = field(default_factory=dict)

    def __call__(self, t: type[T], metadata: dict[str, Any]) -> bool:
        if not issubclass(self.key, t):
            return False

        counter = 0

        for k, v in self.filters.items():
            if k in metadata and metadata[k] == v:
                counter += 1

        return counter >= len(metadata)


class FactoryStorage(MutFactoryRegistry):
    __slots__ = ('_registries',)

    def __init__(self) -> None:
        self._registries = MutDictRegistry[type, Factory[Any]](allow_many=True)

    @override
    def has_factory[T](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        **filters: Any,
    ) -> bool:
        if target_type is None:
            target_type = signature.__args__[-1]

        filter = _CovariantFactoryFilter(key=target_type, filters=filters)

        for _, factory, __ in self._registries.filter(filter):
            if signature is None or factory.check_signature(signature):
                return True

        return False

    @override
    def get_factory[T](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        **filters: Any,
    ) -> Maybe[Factory[T]]:
        if target_type is None:
            target_type = signature.__args__[-1]

        filter = _CovariantFactoryFilter(key=target_type, filters=filters)

        for _, factory, __ in self._registries.filter(filter):
            if signature is None or factory.check_signature(signature):
                return Just(factory)

        return Nothing

    @override
    def get_factory_with_metadata[T](
        self,
        target_type: type[T] | None = None,
        signature: type[Callable] | None = None,
        **filters: Any,
    ) -> Maybe[tuple[Factory[T], Mapping[str, Any]]]:
        if target_type is None:
            target_type = signature.__args__[-1]

        filter = _CovariantFactoryFilter(key=target_type, filters=filters)

        for _, factory, md in self._registries.filter(filter):
            if signature is None or factory.check_signature(signature):
                return Just((factory, md))

        return Nothing

    @override
    def add_factory[T](self, target_type: type[T], factory: Factory[T], **filters: Any) -> bool:
        # filter = _CovariantFactoryFilter(key=target_type, filters=filters)
        # for _ in self._registries.filter(filter):
        #     return False
        return self._registries.add(target_type, factory, **filters)

    @override
    def remove_factory[T](
        self, target_type: type[T], factory: Factory[T] | None = None, **filters: Any
    ) -> Collection[tuple[Factory[T], Mapping[str, Any]]]:
        to_remove: dict[int, tuple[Factory, dict[str, Any]]] = {}

        filter = _CovariantFactoryFilter(key=target_type, filters=filters)
        for _, orig_fn, md in self._registries.filter(filter):
            if factory is None or orig_fn == factory:
                to_remove[id(orig_fn)] = (orig_fn, md)

        res = []

        for orig_fn, md in to_remove.values():
            if self._registries.remove(target_type, orig_fn):
                res.append((orig_fn, md))

        return res



from __future__ import annotations

from app.testmodule.module import __init__

from greyhorse.app.contexts import SyncContext
from greyhorse.app.private.builders.module import ModuleBuilder


def test_module() -> None:
    module_conf = __init__()

    module_builder = ModuleBuilder(module_conf, 'tests', 'test')
    res = module_builder.create_pass()
    assert res.is_ok()
    module = res.unwrap()

    ctx = module.make_facade()

    assert isinstance(ctx, SyncContext)

    with ctx as facade:
        res = facade.add_number(123)
        assert res.is_ok()

        res = facade.get_number()
        assert res.is_ok()
        assert res.unwrap() == 123

        res = facade.remove_number()
        assert res.is_ok()
        assert res.unwrap() is True

        res = facade.get_number()
        assert res.is_err()
        assert res.unwrap_err() == 'Number is not initialized'

    res = module_builder.destroy_pass()
    assert res.is_ok()



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Any

from greyhorse.app.contexts import SyncContext

from ..collections.selectors import Selector
from .resolver import Resolver


class OperationContext(SyncContext[Selector[type, Any]], ABC):
    @property
    @abstractmethod
    def scope[S: IntEnum](self) -> S | None: ...

    @abstractmethod
    def context[S: IntEnum](self, scope: S | None = None) -> OperationContext: ...

    @abstractmethod
    def context_resolver[S: IntEnum](self, scope: S | None = None) -> Resolver: ...

    @abstractmethod
    def advance(self) -> OperationContext: ...



from functools import partial
from typing import Any

import pytest

from greyhorse.run import wrap_async, wrap_sync
from greyhorse.utils.types import is_awaitable


@pytest.fixture(scope='session', autouse=True)
def setup_runtime(request: Any) -> None:
    print('setup_runtime called')

    session = request.node

    for item in session.items:
        match item:
            case pytest.Function() as func:
                if is_awaitable(func.obj):
                    func.obj = partial(wrap_async, func.obj)
                else:
                    func.obj = partial(wrap_sync, func.obj)



from greyhorse.app.schemas.components import ComponentConf, ModuleConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf

from .common.functional import FunctionalOperator
from .components.functional import DictResourceCtrl, FunctionalService, TestLifetime
from .components.resources import DictProviderService


def __init__() -> ModuleConf:  # noqa: N807
    return ModuleConf(
        enabled=True,
        provides=FunctionalOperator,
        fragments=['frag1'],
        components={
            'domain': ComponentConf(enabled=True, services=[SvcConf(type=DictProviderService)]),
            'functional': ComponentConf(
                enabled=True,
                scope_class=TestLifetime,
                controllers=[CtrlConf(type=DictResourceCtrl)],
                services=[SvcConf(type=FunctionalService)],
            ),
        },
    )



from collections.abc import Generator
from copy import deepcopy
from functools import partial
from unittest.mock import Mock

from faker import Faker

from greyhorse.app.abc.providers import BorrowError, BorrowMutError, ForwardError
from greyhorse.app.contexts import (
    MutCtxCallbacks,
    SyncContext,
    SyncMutCallbackContext,
    SyncMutContext,
)
from greyhorse.app.entities.boxes import CombinedBox, ForwardBox, MutBox, SharedBox


def test_shared() -> None:
    instance = SharedBox[int](lambda: 123)

    res = instance.borrow()
    assert res.is_ok()

    res = instance.borrow()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == 123
    instance.reclaim()
    instance.reclaim()


def test_mut() -> None:
    instance = MutBox[int](lambda: 123)

    res = instance.acquire()
    assert res.is_ok()
    unwrapped = res.unwrap()

    res = instance.acquire()
    assert res.is_err()
    assert res.unwrap_err() == BorrowMutError.AlreadyBorrowed(name='int')

    instance.release()

    res = instance.acquire()
    assert res.is_ok()
    assert unwrapped is res.unwrap()

    assert res.unwrap() == 123


def test_combined() -> None:
    instance = CombinedBox[int, str](lambda: 123, lambda: '123')

    res = instance.borrow()
    assert res.is_ok()

    res = instance.borrow()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == 123

    res = instance.acquire()
    assert res.is_err()
    assert res.unwrap_err() == BorrowMutError.BorrowedAsImmutable(name='str')

    instance.reclaim()
    instance.reclaim()

    res = instance.acquire()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == '123'

    res = instance.borrow()
    assert res.is_err()
    assert res.unwrap_err() == BorrowError.BorrowedAsMutable(name='int')

    instance.release()

    res = instance.borrow()
    assert res.is_ok()

    unwrapped = res.unwrap()
    assert unwrapped == 123


def test_shared_context() -> None:
    instance = SharedBox[SyncContext[int]](SyncContext[int](123))

    res = instance.borrow()
    assert res.is_ok()

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == 123


def test_mut_context() -> None:
    instance = MutBox[SyncContext[int]](SyncContext[int](123))

    res = instance.acquire()
    assert res.is_ok()
    unwrapped = res.unwrap()

    res = instance.acquire()
    assert res.is_err()
    assert res.unwrap_err() == BorrowMutError.AlreadyBorrowed(name='IntSyncContext')

    instance.release()

    res = instance.acquire()
    assert res.is_ok()
    assert unwrapped is res.unwrap()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == 123


def test_combined_context() -> None:
    instance = CombinedBox[SyncContext[int], SyncContext[str]](
        SyncContext[int](123), SyncContext[str]('123')
    )

    res = instance.borrow()
    assert res.is_ok()

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == 123

    res = instance.acquire()
    assert res.is_err()
    assert res.unwrap_err() == BorrowMutError.BorrowedAsImmutable(name='StrSyncContext')

    instance.reclaim()
    instance.reclaim()

    res = instance.acquire()
    assert res.is_ok()

    mut_ctx = res.unwrap()
    assert isinstance(mut_ctx, SyncContext)

    with mut_ctx as data:
        assert data == '123'

    res = instance.borrow()
    assert res.is_err()
    assert res.unwrap_err() == BorrowError.BorrowedAsMutable(name='IntSyncContext')

    instance.release()

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == 123


def test_combined_context_read_write() -> None:
    value = {'counter': 1}

    instance = CombinedBox[SyncContext[int], SyncContext[str]](
        SyncContext[dict](partial(deepcopy, value)),
        SyncMutCallbackContext[dict](
            MutCtxCallbacks(on_apply=value.update), partial(deepcopy, value)
        ),
    )

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == {'counter': 1}

    instance.reclaim()

    res = instance.acquire()
    assert res.is_ok()

    mut_ctx = res.unwrap()
    assert isinstance(mut_ctx, SyncMutContext)

    with mut_ctx as data:
        assert data == {'counter': 1}
        data['counter'] += 1

    instance.release()

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == {'counter': 1}

    instance.reclaim()

    res = instance.acquire()
    assert res.is_ok()

    mut_ctx = res.unwrap()
    assert isinstance(mut_ctx, SyncMutContext)

    with mut_ctx as data:
        assert data == {'counter': 1}
        data['counter'] += 1
        mut_ctx.apply()

    instance.release()

    res = instance.borrow()
    assert res.is_ok()

    ctx = res.unwrap()
    assert isinstance(ctx, SyncContext)

    with ctx as data:
        assert data == {'counter': 2}


def test_shared_gen(faker: Faker) -> None:
    gen_mock = Mock()

    def gen() -> Generator[str, str, None]:
        gen_mock()
        yield faker.pystr()
        gen_mock()

    instance = SharedBox[str](gen)

    res = instance.borrow()
    assert res.is_ok()

    gen_mock.assert_called_once()
    gen_mock.reset_mock()

    unwrapped = res.unwrap()
    assert isinstance(unwrapped, str)

    res = instance.borrow()
    assert res.is_ok()

    gen_mock.assert_not_called()
    gen_mock.reset_mock()

    instance.reclaim()
    instance.reclaim()

    assert gen_mock.call_count == 1


def test_mut_gen(faker: Faker) -> None:
    gen_mock = Mock()

    def gen() -> Generator[str, str, None]:
        gen_mock()
        yield faker.pystr()
        gen_mock()

    instance = MutBox[str](gen)

    res = instance.acquire()
    assert res.is_ok()

    gen_mock.assert_called_once()
    gen_mock.reset_mock()

    unwrapped = res.unwrap()
    assert isinstance(unwrapped, str)

    res = instance.acquire()
    assert res.unwrap_err() == BorrowMutError.AlreadyBorrowed(name='str')

    gen_mock.assert_not_called()
    gen_mock.reset_mock()

    instance.release()
    instance.release()

    assert gen_mock.call_count == 1


def test_forward() -> None:
    instance = ForwardBox[int]()

    assert not instance

    res = instance.take()
    assert res.is_err()
    assert res.unwrap_err() == ForwardError.Empty(name='int')

    instance.accept(123)

    assert instance

    res = instance.take()
    assert res.is_ok()

    assert not instance

    unwrapped = res.unwrap()
    assert unwrapped == 123

    res = instance.take()
    assert res.is_err()
    assert res.unwrap_err() == ForwardError.Empty(name='int')

    instance.drop(unwrapped)



from abc import ABC, abstractmethod
from collections.abc import Collection, Mapping
from typing import Any

from greyhorse.factory import Factory

from .functional.context import OperationContext
from .resources.operators import MutOperatorCollector, OperatorCollector
from .resources.providers import MutProviderCollector, ProviderCollector


class Component(ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def context(self) -> OperationContext: ...

    @abstractmethod
    def add_factory[T](
        self, target_type: type[T], factory: Factory[T], **filters: Any
    ) -> bool: ...

    @abstractmethod
    def remove_factory[T](
        self, target_type: type[T], factory: Factory[T] | None = None, **filters: Any
    ) -> Collection[tuple[Factory[T], Mapping[str, Any]]]: ...

    @abstractmethod
    def setup(
        self,
        op_collector: OperatorCollector | None = None,
        prov_collector: ProviderCollector | None = None,
    ) -> None: ...

    @abstractmethod
    def teardown(
        self,
        op_collector: MutOperatorCollector | None = None,
        prov_collector: MutProviderCollector | None = None,
    ) -> None: ...



from copy import deepcopy
from functools import partial
from typing import override

from greyhorse.app.abc.resources.providers import provider
from greyhorse.app.abc.services import ServiceState
from greyhorse.app.contexts import MutCtxCallbacks, SyncContext, SyncMutCallbackContext
from greyhorse.app.entities.boxes import OwnerCtxRefBox
from greyhorse.app.entities.services import SyncService
from greyhorse.factory import CachePolicy

from ..common.resources import DictResContext, DictResource, MutDictResContext


class DictResourceBox(OwnerCtxRefBox[DictResource, DictResource]):
    allow_borrow_when_acquired = True
    allow_acq_when_borrowed = True
    allow_multiple_acquisition = False


class DictProviderService(SyncService):
    def __init__(self) -> None:
        super().__init__()
        self._value = {}
        self._box = DictResourceBox(
            SyncContext,
            SyncMutCallbackContext,
            partial(deepcopy, self._value),
            partial(deepcopy, self._value),
            mut_params=dict(callbacks=MutCtxCallbacks(on_apply=self._setter)),
        )

    def _setter(self, value: DictResource) -> None:
        self._value.clear()
        self._value.update(value)

    @override
    def setup(self) -> None:
        self._set_state(ServiceState.Ready(started=False))

    @override
    def teardown(self) -> None:
        self._set_state(ServiceState.Unready(started=False, message='stopped'))

    @provider(policy=CachePolicy.VOLATILE)
    def create_dict(self) -> DictResContext:
        value = self._box.borrow().unwrap()
        yield value
        self._box.reclaim(value)

    @provider()
    def create_mut_dict(self) -> MutDictResContext:
        value = self._box.acquire().unwrap()
        yield value
        self._box.release(value)



from __future__ import annotations

from collections.abc import AsyncGenerator, Generator
from decimal import Decimal
from enum import IntEnum, auto

from greyhorse.app.private.functional.context import BucketOperationContext
from greyhorse.app.private.functional.factory_storage import FactoryStorage
from greyhorse.factory import into_factory


def sync_ctx_manager() -> Generator[str, str, None]:
    res = '123'
    yield res
    assert res == '123'


async def async_ctx_manager() -> AsyncGenerator[Decimal, Decimal]:
    res = Decimal(789)
    yield res
    assert res == Decimal(789)


def complex_fn(a: int, b: str) -> Decimal:
    return Decimal(a) + Decimal(b)


class Level(IntEnum):
    ROOT = auto(0)
    FUNCTION = auto()
    LOCAL = auto()


def test_simple() -> None:
    registry = FactoryStorage()

    context = BucketOperationContext(registry)
    assert context.scope is None

    with context as selector:
        assert not selector.has(int)
        assert not selector.has(str)
        assert not selector.has(Decimal)

        assert registry.add_factory(int, into_factory(lambda: 123, int))
        assert registry.add_factory(str, into_factory(sync_ctx_manager))

        assert selector.has(int)
        assert selector.has(str)
        assert not selector.has(Decimal)

        assert selector.get(int).unwrap() == 123

        inner_context = context.advance()
        assert context is inner_context


def test_scoped() -> None:
    registry = FactoryStorage()

    context = BucketOperationContext(registry, Level)
    assert context.scope == Level.ROOT

    with context as selector:
        assert not selector.has(int)
        assert not selector.has(str)
        assert not selector.has(Decimal)

        assert registry.add_factory(int, into_factory(lambda: 123, int), scope=Level.ROOT)
        assert registry.add_factory(str, into_factory(sync_ctx_manager), scope=Level.ROOT)

        assert selector.has(int)
        assert selector.has(str)
        assert not selector.has(Decimal)

        assert selector.get(int).unwrap() == 123

        inner_context = context.advance()
        assert context is not inner_context

        with inner_context as inner_selector:
            assert inner_context.scope == Level.FUNCTION
            assert registry.add_factory(
                Decimal, into_factory(async_ctx_manager), scope=Level.FUNCTION
            )

            assert inner_selector.has(int)
            assert inner_selector.has(str)
            assert inner_selector.has(Decimal)

            assert registry.add_factory(int, into_factory(lambda: 456, int), scope=Level.LOCAL)

            inner_context2 = inner_context.advance()
            assert inner_context is not inner_context2

            with inner_context2 as inner_selector2:
                assert inner_context2.scope == Level.LOCAL
                assert inner_selector2.get(int).unwrap() == 456
                assert inner_selector2.get(str).unwrap() == '123'
                assert Decimal(789) == inner_selector2.get(Decimal).unwrap()

            with inner_context2 as inner_selector2:
                assert registry.add_factory(
                    Decimal, into_factory(complex_fn), scope=Level.LOCAL
                )
                assert Decimal(579) == inner_selector2.get(Decimal).unwrap()

        assert inner_selector.get(int).unwrap() == 123
        assert registry.remove_factory(Decimal, scope=Level.FUNCTION)

        assert registry.remove_factory(int, scope=Level.ROOT)

    with context as selector:
        assert not selector.has(int)
        assert selector.has(str)
        assert not selector.has(Decimal)



from abc import ABC, abstractmethod
from collections.abc import Callable, Collection, Mapping
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Any

from greyhorse.factory import Factory

from ..collections.collectors import Collector, MutCollector
from .common import ResourceEventKind


@dataclass(slots=True, frozen=True, kw_only=True)
class OperatorData:
    component: str
    fragment: str
    target: Factory[Any]
    signature: type[Callable]
    scope: IntEnum | None = None
    external_params: Collection[type] | None = None
    metadata: Mapping[str, Any] = field(default_factory=dict, compare=False)


OperatorCollector = Collector[type, OperatorData]
MutOperatorCollector = MutCollector[type, OperatorData]


@dataclass(slots=True, frozen=True, kw_only=True)
class OperatorEvent:
    kind: ResourceEventKind
    type: type
    data: OperatorData


type OperatorListenerFn = Callable[[OperatorEvent], ...]


class OperatorSubscription(ABC):
    @abstractmethod
    def add_operator_listener[T](
        self,
        key: type[T],
        fn: OperatorListenerFn,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool: ...

    @abstractmethod
    def remove_operator_listener[T](
        self,
        key: type[T],
        fn: OperatorListenerFn | None = None,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool: ...



from .private.builders.module import ModuleBuilder
from .private.component import Component
from .private.controllers import Controller, operator_listener
from .private.fragment import Fragment, factory, operator, provider
from .private.module import Module
from .private.schemas.component import ComponentConf, ModuleComponentConf, ModuleConf
from .private.schemas.ctrl import CtrlConf
from .private.schemas.svc import SvcConf
from .private.services import Service, provider_listener



import enum
from datetime import datetime
from decimal import Decimal
from typing import Any, override

import pytest
from broken.controllers import Condition, ControllerState

from greyhorse.app.abc.providers import ProviderSelector
from greyhorse.app.abc.resources.common import ResourceEventKind
from greyhorse.app.abc.resources.operators import OperatorEvent, operator, operator_listener
from greyhorse.app.abc.resources.providers import (
    ProviderEvent,
    mut_provider,
    provider,
    provider_listener,
)
from greyhorse.app.abc.services import ServiceState
from greyhorse.app.contexts import SyncContext, SyncMutContext
from greyhorse.app.entities.component import AsyncComponent, ComponentState, SyncComponent
from greyhorse.app.entities.controllers import AsyncController, SyncController
from greyhorse.app.entities.fragment import Fragment
from greyhorse.app.entities.module import Module
from greyhorse.app.entities.services import AsyncService, SyncService
from greyhorse.factory import into_factory
from greyhorse.maybe import Maybe
from greyhorse.result import Ok, Result


class TestLifetime(enum.IntEnum):
    DOMAIN = enum.auto(0)
    OPERATION = enum.auto()


DictResource = dict[str, Any]
DictResContext = SyncContext[DictResource]
MutDictResContext = SyncMutContext[DictResource]


class FunctionalOperator:
    def __init__(self, ctx: DictResContext, mut_ctx: MutDictResContext) -> None:
        self.ctx = ctx
        self.mut_ctx = mut_ctx

    def add_number(self, value: int) -> Result[None, str]:
        with self.mut_ctx as ctx:
            ctx['number'] = value
            self.mut_ctx.apply()

        return Ok()

    def get_number(self) -> Result[int, str]:
        with self.ctx as ctx:
            value = Maybe(ctx.get('number'))

        return value.ok_or('Number is not initialized')

    def remove_number(self) -> Result[bool, str]:
        with self.mut_ctx as ctx:
            value = Maybe(ctx.pop('number', None))
            self.mut_ctx.apply()

        return value.ok_or('Number is not initialized').map(lambda _: True)


class SyncTestService(SyncService):
    def __init__(self) -> None:
        super().__init__()
        self._res = {}
        self._ctx = SyncContext[DictResource](lambda: self._res)
        self._mut_ctx = SyncMutContext[DictResource](lambda: self._res)

    @override
    def setup(self) -> None:
        self._set_state(ServiceState.Ready(started=False))

    @override
    def teardown(self) -> None:
        self._set_state(ServiceState.Unready(started=False, message='stopped'))

    @provider()
    def p_1(self) -> DictResContext:
        return self._ctx

    @mut_provider()
    def p_2(self) -> MutDictResContext:
        return self._mut_ctx

    @provider(scope=TestLifetime.OPERATION)
    def prov_1(self, global_dict: dict) -> int:
        global_dict['SyncTestService'] = 123
        yield 123
        del global_dict['SyncTestService']

    @operator(scope=TestLifetime.DOMAIN, fragment='frag1')
    def op_1(self, global_dict: dict) -> float:
        global_dict['test'] = 123
        yield global_dict['res'] + 1000
        del global_dict['test']


class AsyncTestService(AsyncService):
    def __init__(self) -> None:
        super().__init__()

    @override
    async def setup(self) -> None:
        self._set_state(ServiceState.Ready(started=False))

    @override
    async def teardown(self) -> None:
        self._set_state(ServiceState.Unready(started=False, message='stopped'))

    f_2 = provider(scope=TestLifetime.OPERATION, fragment='frag1')(FunctionalOperator)
    o_1 = operator(scope=TestLifetime.OPERATION, fragment='frag1')(FunctionalOperator)

    @provider(scope=TestLifetime.OPERATION, fragment='frag1')
    async def prov_1(self, global_dict: dict) -> Decimal:
        global_dict['AsyncTestService'] = 123
        yield Decimal(123)
        del global_dict['AsyncTestService']

    @operator(scope=TestLifetime.DOMAIN)
    async def op_1(self, global_dict: dict, e: str) -> str:
        global_dict['test'] = 123
        yield str(global_dict['res'] + 1000)
        del global_dict['test']


class SyncResourceCtrl(SyncController):
    def __init__(self, providers: ProviderSelector) -> None:
        super().__init__()
        self._providers = providers

    @override
    def setup(self) -> None:
        cond = Condition(
            name='DictResContext',
            status=self._providers.has(DictResContext),
            permanent=True,
            transition_dt=datetime.now(),
        )
        self._set_condition(cond)
        cond = Condition(
            name='MutDictResContext',
            status=self._providers.has(MutDictResContext),
            permanent=True,
            transition_dt=datetime.now(),
        )
        self._set_condition(cond)
        self._set_state(ControllerState.Succeeded)

    @override
    def teardown(self) -> None:
        self._reset_condition('DictResContext')
        self._reset_condition('MutDictResContext')
        self._set_state(ControllerState.Finalized)

    @provider_listener(DictResContext, fragment='frag1')
    def listener_1(self, event: ProviderEvent) -> None:
        if event.kind == ResourceEventKind.SETUP:
            f = event.context.__enter__()
            self._factories.add_factory(DictResContext, f)
        elif event.kind == ResourceEventKind.TEARDOWN:
            self._factories.remove_factory(DictResContext)
            event.context.__exit__()

    @provider_listener(MutDictResContext, fragment='frag1')
    def listener_2(self, event: ProviderEvent) -> None:
        if event.kind == ResourceEventKind.SETUP:
            f = event.context.__enter__()
            self._factories.add_factory(MutDictResContext, f)
        elif event.kind == ResourceEventKind.TEARDOWN:
            self._factories.remove_factory(MutDictResContext)
            event.context.__exit__()

    @operator_listener(float | str, fragment='frag1')
    def listener_3(self, event: OperatorEvent) -> None:
        pass


class AsyncResourceCtrl(AsyncController):
    def __init__(self) -> None:
        super().__init__()

    @override
    async def setup(self, global_dict: dict) -> None:
        cond = Condition(name='dict', status=False, permanent=True, message='qwer')
        self._set_condition(cond)
        self._set_state(ControllerState.Running(alive=True))

    @override
    async def teardown(self, global_dict: dict) -> None:
        self._reset_condition('dict')
        self._set_state(ControllerState.Succeeded)

    @provider_listener(int | Decimal, fragment='frag1')
    async def listener_1(self, event: ProviderEvent) -> None:
        pass

    @operator_listener(float | str, fragment='frag1')
    async def listener_2(self, event: OperatorEvent) -> None:
        pass


def test_sync_component() -> None:
    state = ComponentState(name='test')
    component = SyncComponent(state)

    assert component.add_controller(SyncResourceCtrl)
    assert not component.add_controller(AsyncResourceCtrl)
    assert component.add_service(SyncTestService)
    assert not component.add_service(AsyncTestService)

    fragment = Fragment('test', 'frag1', scope_class=TestLifetime)

    global_dict = {'res': 345}
    assert fragment.add_operator(dict, into_factory(global_dict))

    component.register_controllers({fragment.name: fragment})
    component.register_services({fragment.name: fragment})

    with fragment.context as selector:
        assert selector.has(float)
        value = selector.get(float).unwrap()
        assert value == 1345
        assert global_dict == {'res': 345, 'test': 123}

    assert global_dict == {'res': 345}

    component.setup_providers()

    res = component.get_shared_provider(int)
    assert res.is_just()
    prov = res.unwrap()

    res = prov.borrow()
    assert res.is_ok()
    assert res.unwrap() == 123
    assert global_dict == {'res': 345, 'SyncTestService': 123}

    prov.reclaim()
    assert global_dict == {'res': 345}

    res = component.get_shared_provider(DictResContext)
    assert res.is_just()
    prov = res.unwrap()
    res = prov.borrow()
    assert res.is_ok()
    prov.reclaim()

    res = component.get_mut_provider(MutDictResContext)
    assert res.is_just()
    prov = res.unwrap()
    res = prov.acquire()
    assert res.is_ok()
    prov.release()

    component.teardown_providers()

    component.unregister_controllers({fragment.name: fragment})
    component.unregister_services({fragment.name: fragment})

    assert component.remove_service(SyncTestService)
    assert not component.remove_service(AsyncTestService)
    assert component.remove_controller(SyncResourceCtrl)
    assert not component.remove_controller(AsyncResourceCtrl)


@pytest.mark.asyncio
async def test_async_component() -> None:
    component = AsyncComponent('test', 'name', TestLifetime)

    assert component.add_controller(SyncResourceCtrl)
    assert component.add_controller(AsyncResourceCtrl)
    assert component.add_service(SyncTestService)
    assert component.add_service(AsyncTestService)

    fragment = Fragment('test', 'frag1')

    global_dict = {'res': 345}
    assert component.add_factory(dict, into_factory(global_dict))

    component.register_controllers({fragment.name: fragment})
    component.register_services({fragment.name: fragment})

    op = fragment._registry.get_factory(float).unwrap()
    value = op()
    assert value == 1345
    assert global_dict == {'res': 345, 'test': 123}

    op.destroy(value)
    assert global_dict == {'res': 345}

    op = fragment._registry.get_factory(str).unwrap()
    value = op()
    assert value == '1345'
    assert global_dict == {'res': 345, 'test': 123}

    op.destroy(value)
    assert global_dict == {'res': 345}

    prov_ctx = fragment._prov_contexts.get(int).unwrap()
    with prov_ctx as prov:
        value = prov()
        assert value == 123
        assert global_dict == {'res': 345, 'SyncTestService': 123}

        prov.destroy(value)
        assert global_dict == {'res': 345}

    prov_ctx = fragment._prov_contexts.get(Decimal).unwrap()
    with prov_ctx as prov:
        value = await prov()
        assert Decimal(123) == value
        assert global_dict == {'res': 345, 'AsyncTestService': 123}

        await prov.destroy(value)
        assert global_dict == {'res': 345}

    component.unregister_controllers({fragment.name: fragment})
    component.unregister_services({fragment.name: fragment})

    assert component.remove_service(SyncTestService)
    assert component.remove_service(AsyncTestService)
    assert component.remove_controller(SyncResourceCtrl)
    assert component.remove_controller(AsyncResourceCtrl)


@pytest.mark.asyncio
async def test_async_two_components() -> None:
    component1 = AsyncComponent('test', '1', TestLifetime)
    component2 = AsyncComponent('test', '2', TestLifetime)

    assert component2.add_controller(SyncResourceCtrl)
    # assert component2.add_controller(AsyncResourceCtrl)
    assert component1.add_service(SyncTestService)
    assert component2.add_service(AsyncTestService)

    fragment = Fragment('test', 'frag1')

    component1.register_controllers({fragment.name: fragment})
    component2.register_controllers({fragment.name: fragment})
    component1.register_services({fragment.name: fragment})
    component2.register_services({fragment.name: fragment})

    prov_ctx = fragment._prov_contexts.get(FunctionalOperator).unwrap()
    with prov_ctx as prov:
        oper = prov()

        res = oper.add_number(123)
        assert res.is_ok()

        res = oper.get_number()
        assert res.is_ok()
        assert res.unwrap() == 123

        res = oper.remove_number()
        assert res.is_ok()
        assert res.unwrap() is True

        res = oper.get_number()
        assert res.is_err()
        assert res.unwrap_err() == 'Number is not initialized'

        prov.destroy(oper)

    component2.unregister_services({fragment.name: fragment})
    component1.unregister_services({fragment.name: fragment})
    component2.unregister_controllers({fragment.name: fragment})
    component1.unregister_controllers({fragment.name: fragment})

    assert component1.remove_service(SyncTestService)
    assert component2.remove_service(AsyncTestService)
    assert component2.remove_controller(SyncResourceCtrl)
    # assert component2.remove_controller(AsyncResourceCtrl)


@pytest.mark.asyncio
async def test_async_module() -> None:
    component1 = AsyncComponent('test', '1', TestLifetime)
    component2 = AsyncComponent('test', '2', TestLifetime)

    assert component2.add_controller(SyncResourceCtrl)
    # assert component2.add_controller(AsyncResourceCtrl)
    assert component1.add_service(SyncTestService)
    assert component2.add_service(AsyncTestService)

    module = Module[FunctionalOperator](
        'test',
        'test',
        FunctionalOperator,
        components=[component1, component2],
        fragments=['frag1'],
    )

    ctx = module.make_facade()

    async with ctx as facade:
        res = facade.add_number(123)
        assert res.is_ok()

        res = facade.get_number()
        assert res.is_ok()
        assert res.unwrap() == 123

        res = facade.remove_number()
        assert res.is_ok()
        assert res.unwrap() is True

        res = facade.get_number()
        assert res.is_err()
        assert res.unwrap_err() == 'Number is not initialized'

    assert component1.remove_service(SyncTestService)
    assert component2.remove_service(AsyncTestService)
    assert component2.remove_controller(SyncResourceCtrl)
    # assert component2.remove_controller(AsyncResourceCtrl)



import asyncio
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable


class Runtime(ABC):
    __slots__ = ('_loop',)

    _instance: 'Runtime' = None

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        self._loop = loop

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    @abstractmethod
    def active(self) -> bool: ...

    @abstractmethod
    def start(self) -> None | Awaitable[None]: ...

    @abstractmethod
    def stop(self) -> None | Awaitable[None]: ...

    @abstractmethod
    def invoke_sync[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T: ...

    @abstractmethod
    async def invoke_async[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T: ...



from abc import ABC, abstractmethod
from collections.abc import Collection
from dataclasses import dataclass, field

from greyhorse.factory import Factory


@dataclass(slots=True, frozen=True, kw_only=True)
class CompilationResult:
    resolved: Collection[type] = field(default_factory=list)
    unresolved: Collection[type] = field(default_factory=list)

    @property
    def success(self) -> bool:
        return self.resolved and not self.unresolved

    @property
    def partial(self) -> bool:
        return bool(self.resolved) and bool(self.unresolved)

    @property
    def failed(self) -> bool:
        return not self.resolved and self.unresolved


class Operator[T](ABC):
    @property
    @abstractmethod
    def return_type(self) -> type[T]: ...

    @property
    @abstractmethod
    def compiled(self) -> bool: ...

    @abstractmethod
    def compile(self) -> CompilationResult: ...

    @abstractmethod
    def get_functor(self) -> Factory[T]: ...



from greyhorse.app.entities.fragment import Fragment
from greyhorse.factory import into_factory


def test_fragment_factory() -> None:
    fragment = Fragment('test', 'name')

    assert not fragment.has_factory(int)
    assert fragment.get_factory(int).is_nothing()

    assert fragment.add_factory(int, into_factory(lambda: 123, int))

    assert fragment.has_factory(int)
    assert fragment.get_factory(int).is_just()

    f = fragment.get_factory(int).unwrap()
    assert f() == 123

    assert fragment.remove_factory(int)

    assert not fragment.has_factory(int)
    assert fragment.get_factory(int).is_nothing()


def test_fragment_context() -> None:
    fragment = Fragment('test', 'name')

    assert not fragment.has_factory(int)
    assert fragment.get_factory(int).is_nothing()

    assert fragment.add_factory(int, into_factory(lambda: 123, int))

    assert fragment.has_factory(int)
    assert fragment.get_factory(int).is_just()

    with fragment.get_context() as selector:
        assert selector.has(int)
        assert selector.get(int).unwrap() == 123

    assert fragment.remove_factory(int)

    assert not fragment.has_factory(int)
    assert fragment.get_factory(int).is_nothing()



from greyhorse.error import Error, ErrorCase


class BorrowError(Error):
    namespace = 'greyhorse.app'

    Empty = ErrorCase(
        msg='Cannot borrow "{name}" as immutable because the value is not available', name=str
    )

    # MovedOut = ErrorCase(
    #     msg='Cannot borrow "{name}" as immutable because the value was moved out', name=str
    # )

    BorrowedAsMutable = ErrorCase(
        msg='Cannot borrow "{name}" as immutable because it is also borrowed as mutable',
        name=str,
    )

    Unexpected = ErrorCase(
        msg='Cannot borrow "{name}" as immutable because an unexpected '
        'error occurred: "{details}"',
        name=str,
        details=str,
    )


class BorrowMutError(Error):
    namespace = 'greyhorse.app'

    Empty = ErrorCase(
        msg='Cannot borrow "{name}" as mutable because the value is not available', name=str
    )

    # MovedOut = ErrorCase(
    #     msg='Cannot borrow "{name}" as mutable because the value was moved out', name=str
    # )

    AlreadyBorrowed = ErrorCase(
        msg='Cannot borrow "{name}" as mutable more than once at a time', name=str
    )

    BorrowedAsImmutable = ErrorCase(
        msg='Cannot borrow "{name}" as mutable because it is also borrowed as immutable',
        name=str,
    )

    Unexpected = ErrorCase(
        msg='Cannot borrow "{name}" as mutable because an unexpected '
        'error occurred: "{details}"',
        name=str,
        details=str,
    )



from typing import Any

from pydantic import AliasChoices, BaseModel, Field, PrivateAttr, model_validator

from ..controllers import Controller
from ..runtime.invoke import caller_path


class CtrlConf(BaseModel, frozen=True):
    type_: type[Controller] = Field(validation_alias=AliasChoices('type'))
    name: str | None = None
    args: dict[str, Any] = Field(default_factory=dict)
    enabled: bool = Field(default=True)
    _init_path: list[str] = PrivateAttr(default_factory=lambda: caller_path(5))

    @model_validator(mode='before')
    def _setup_name(self: dict[str, Any]) -> dict:
        if 'name' not in self:
            self['name'] = self['type'].__name__
        return self

    @property
    def init_path(self) -> list[str]:
        return self._init_path.copy()



from abc import ABC, abstractmethod

from greyhorse.maybe import Maybe

from .component import Component


class Module(ABC):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def full_path(self) -> str: ...

    @abstractmethod
    def get_component(self, name: str) -> Maybe[Component]: ...

    @abstractmethod
    def setup(self) -> None: ...
    @abstractmethod
    def teardown(self) -> None: ...



from __future__ import annotations

from collections.abc import Collection
from typing import Any, override

from greyhorse.app.abc.component import Component
from greyhorse.app.abc.module import Module as BaseModule
from greyhorse.app.abc.resources.reconciler import (
    ReconcilerCollector,
    ReconcilerRunPhase,
    ReconcilerTaskFn,
    RestartPolicy,
)
from greyhorse.app.abc.resources.state import (
    MutResourceSelector,
    ResourceSelector,
    ResourceState,
)
from greyhorse.logging import logger
from greyhorse.maybe import Maybe
from greyhorse.utils.injectors import ParamsInjector

from .collections.registries import MutDictRegistry
from .controllers import Controller
from .fragment import Fragment
from .resources.registries import OperatorRegistry, ProviderRegistry
from .services import Service


class _ReconcilerCollector(ReconcilerCollector):
    def __init__(self) -> None: ...

    @override
    def add_task[T](
        self,
        key: type[T],
        function: ReconcilerTaskFn[T],
        phase: ReconcilerRunPhase = ReconcilerRunPhase.Setup,
        restart_policy: RestartPolicy = RestartPolicy.Never,
        conditions: list[str] | None = None,
    ) -> bool: ...

    @override
    def remove_task[T](
        self,
        key: type[T],
        function: ReconcilerTaskFn[T],
        phase: ReconcilerRunPhase = ReconcilerRunPhase.Setup,
    ) -> bool: ...


class Module(BaseModule):
    __slots__ = (
        '_components',
        '_controllers',
        '_fragments',
        '_injector',
        '_name',
        '_op_registry',
        '_parent',
        '_prov_registry',
        '_resources',
        '_services',
    )

    def __init__(
        self,
        name: str,
        parent: Module | None = None,
        fragments: Collection[Fragment] | None = None,
        components: Collection[Component] | None = None,
    ) -> None:
        self._name = name
        self._parent = parent
        self._services = MutDictRegistry[str, Service]()
        self._controllers = MutDictRegistry[str, Controller]()
        self._resources = MutDictRegistry[str, ResourceState]()
        self._injector = ParamsInjector()
        self._prov_registry = ProviderRegistry()
        self._op_registry = OperatorRegistry()
        self._fragments = {f.name: f for f in fragments} if fragments else {}
        self._components = {c.name: c for c in components} if components else {}

    @property
    @override
    def name(self) -> str:
        return self._name

    @property
    @override
    def full_path(self) -> str:
        if self._parent is None:
            return self._name
        return f'{self._parent.full_path}.{self._name}'

    def __repr__(self) -> str:
        return f'Module "{self.full_path}"'

    @override
    def get_component(self, name: str) -> Maybe[Component]:
        return Maybe(self._components.get(name))

    @override
    def setup(self) -> None:
        for _name, ctrl in self._controllers.list():
            ctrl.setup(self._op_registry, None)  # TODO
        for _name, svc in self._services.list():
            svc.setup(self._prov_registry, None)  # TODO
        for comp in self._components.values():
            comp.setup(self._op_registry, self._prov_registry)

    @override
    def teardown(self) -> None:
        for comp in reversed(self._components.values()):
            comp.teardown(self._op_registry, self._prov_registry)
        for _name, svc in reversed(list(self._services.list())):
            svc.teardown(self._prov_registry)
        for _name, ctrl in reversed(list(self._controllers.list())):
            ctrl.teardown(self._op_registry)

    def add_service(
        self,
        type_: type[Service],
        name: str | None = None,
        init_path: list[str] | None = None,
        **kwargs: Any,
    ) -> bool:
        if not name:
            name = type_.__name__

        if self._services.has(name):
            return False

        logger.info('{path}: Service "{name}" create'.format(path=self.full_path, name=name))

        factory = type_
        values = kwargs.copy()
        values['name'] = name

        if init_path:
            values['init_path'] = init_path

        types = {ResourceSelector: self._resources}
        injected = self._injector(factory, values, types)
        instance = factory(*injected.args, **injected.kwargs)

        logger.info('{path}: Service "{name}" created'.format(path=self.full_path, name=name))
        return self._services.add(name, instance)

    def add_controller(
        self,
        type_: type[Controller],
        name: str | None = None,
        init_path: list[str] | None = None,
        **kwargs: Any,
    ) -> bool:
        if not name:
            name = type_.__name__

        if self._controllers.has(name):
            return False

        logger.info('{path}: Controller "{name}" create'.format(path=self.full_path, name=name))

        factory = type_
        values = kwargs.copy()
        values['name'] = name

        if init_path:
            values['init_path'] = init_path

        types = {MutResourceSelector: self._resources}
        injected = self._injector(factory, values, types)
        instance = factory(*injected.args, **injected.kwargs)

        logger.info(
            '{path}: Controller "{name}" created'.format(path=self.full_path, name=name)
        )

        return self._controllers.add(name, instance)

    def remove_service(self, type_: type[Service], name: str | None = None) -> bool:
        if not name:
            name = type_.__name__

        if not self._services.has(name):
            return False

        instance = self._services.get(name).unwrap()
        return self._services.remove(name, instance)

    def remove_controller(self, type_: type[Controller], name: str | None = None) -> bool:
        if not name:
            name = type_.__name__

        if not self._controllers.has(name):
            return False

        instance = self._controllers.get(name).unwrap()
        return self._controllers.remove(name, instance)



from abc import ABC, abstractmethod
from collections.abc import Callable, Mapping
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Any

from ..collections.collectors import Collector, MutCollector
from .common import ResourceEventKind


@dataclass(slots=True, frozen=True, kw_only=True)
class ProviderData:
    component: str
    fragment: str
    signature: type[Callable]
    scope: IntEnum | None = None
    metadata: Mapping[str, Any] = field(default_factory=dict, compare=False)


ProviderCollector = Collector[type, ProviderData]
MutProviderCollector = MutCollector[type, ProviderData]


@dataclass(slots=True, frozen=True, kw_only=True)
class ProviderEvent:
    kind: ResourceEventKind
    type: type
    data: ProviderData


type ProviderListenerFn = Callable[[ProviderEvent], ...]


class ProviderSubscription(ABC):
    @abstractmethod
    def add_provider_listener[T](
        self,
        key: type[T],
        fn: ProviderListenerFn,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool: ...

    @abstractmethod
    def remove_provider_listener[T](
        self,
        key: type[T],
        fn: ProviderListenerFn | None = None,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool: ...



from __future__ import annotations

from collections.abc import Collection, Generator
from enum import IntEnum
from typing import Any, override

from greyhorse.app.abc.functional.context import OperationContext
from greyhorse.app.abc.functional.operators import CompilationResult, Operator
from greyhorse.app.abc.functional.resolver import ResolvedData, Resolver
from greyhorse.factory import Factory, into_factory


class ContextOperator[T](Operator[T]):
    __slots__ = (
        '_compiled',
        '_external_params',
        '_functor',
        '_resolved_types',
        '_resolver',
        '_scope',
    )

    def __init__(
        self,
        functor: Factory[T],
        resolver: Resolver,
        scope: IntEnum | None = None,
        external_params: Collection[type] | None = None,
    ) -> None:
        self._resolver = resolver
        self._scope = scope
        self._external_params = set(external_params or [])
        self._resolved_types: dict[type, ResolvedData] = {
            functor.return_type: ResolvedData(
                type=functor.return_type, factory=functor, deps=dict(functor.actual_params)
            )
        }
        self._compiled = False
        self._functor = functor

    @override
    @property
    def return_type(self) -> type[T]:
        return self._functor.return_type

    @override
    @property
    def compiled(self) -> bool:
        return self._compiled

    @classmethod
    def from_context(
        cls,
        context: OperationContext,
        functor: Factory[T],
        scope: IntEnum | None,
        external_params: Collection[type] | None = None,
    ) -> Operator[T]:
        resolver = context.context_resolver(scope)
        return cls(functor, resolver, scope, external_params)

    @override
    def compile(self) -> CompilationResult:
        if self.compiled:
            return CompilationResult(resolved=self._resolved_types.keys())

        res = self._resolver.resolve_factories(self.return_type, self._scope)
        self._resolved_types.update(res.resolved)
        self._compiled = not res.unresolved
        return CompilationResult(
            resolved=self._resolved_types.keys(), unresolved=res.unresolved
        )

    @override
    def get_functor(self) -> Factory[T]:
        compilation_result = self.compile()

        external_params = {
            param_name: param_type
            for param_name, param_type in self._functor.actual_params.items()
            if param_type in self._external_params
            or param_type in compilation_result.unresolved
            or not self._resolver.can_resolve(param_type)
        }

        external_params_size = len(external_params) if external_params else 0
        if external_params_size >= len(self._functor.actual_params):
            return self._functor

        return Factory[self.return_type].from_syncgen(
            self._entrypoint,
            cache_policy=self._functor.cache_policy,
            hints={
                'return': Generator[self.return_type, self.return_type, None],
                **external_params,
            },
        )

    def _entrypoint(self, *args: Any, **kwargs: Any) -> Generator[T, T, None]:
        resolved = self._resolved_types.copy()

        with self._resolver as resolver:
            for param_idx, (param_name, param_type) in enumerate(
                self._functor.actual_params.items()
            ):
                if param_idx < len(args) and issubclass(type(args[param_idx]), param_type):
                    resolved[param_type] = ResolvedData(
                        type=param_type, factory=into_factory(args[param_idx])
                    )
                    continue
                if param_name in kwargs and param_type in self._external_params:
                    resolved[param_type] = ResolvedData(
                        type=param_type, factory=into_factory(kwargs[param_name])
                    )
                    continue
            instance = resolver.resolve_value(resolved, self.return_type, self._scope)
            assert instance.is_just()
            yield instance.unwrap()



from collections.abc import AsyncGenerator, Callable, Generator
from decimal import Decimal
from unittest.mock import Mock

import pytest
from faker import Faker

from greyhorse.factory import into_factory
from greyhorse.maybe import Maybe


def sync_simple_fn() -> int:
    return 123


async def async_simple_fn() -> int:
    return 123


def sync_complex_fn(a1: int, a2: str) -> Decimal:
    return Decimal(a1) + Decimal(a2)


async def async_complex_fn(a1: int, a2: str) -> Decimal:
    return Decimal(a1) + Decimal(a2)


def sync_ctx_fn(a1: int, a2: str) -> Generator[Decimal, Decimal, None]:
    res = Decimal(a1) + Decimal(a2)
    yield res
    assert res == Decimal(a1) + Decimal(a2)


async def async_ctx_fn(a1: int, a2: str) -> AsyncGenerator[Decimal, Decimal]:
    res = Decimal(a1) + Decimal(a2)
    yield res
    assert res == Decimal(a1) + Decimal(a2)


class A:
    pass


class B:
    def __init__(self, a: A) -> None:
        self.a = a


class C:
    def __init__(self, conf: str) -> None:
        self.conf = conf


class D:
    def __init__(self, a: A, b: Maybe[B], c: C | None) -> None:
        self.a = a
        self.b = b
        self.c = c


def test_instance() -> None:
    factory = into_factory('test')

    assert not factory.scoped
    assert factory.cacheable
    assert not factory.is_async

    assert factory.return_type is str
    assert factory.params_count == 0
    assert factory.args_count == 0
    assert factory.actual_params == {}

    assert factory.check_signature(Callable[[], str])
    assert factory.check_signature(Callable[[], object])

    res = factory.create()
    assert isinstance(res, str)
    assert res == 'test'


def test_classes() -> None:
    factory_a = into_factory(A)

    assert not factory_a.scoped
    assert factory_a.cacheable
    assert not factory_a.is_async

    assert factory_a.return_type is A
    assert factory_a.params_count == 0
    assert factory_a.args_count == 0
    assert factory_a.actual_params == {}

    assert factory_a.check_signature(Callable[[], A])
    assert factory_a.check_signature(Callable[[], object])

    res_a = factory_a.create()
    assert isinstance(res_a, A)

    factory_b = into_factory(B)

    assert not factory_b.scoped
    assert factory_b.cacheable
    assert not factory_b.is_async

    assert factory_b.return_type is B
    assert factory_b.params_count == 1
    assert factory_b.args_count == 0
    assert factory_b.actual_params == {'a': A}

    assert factory_b.has_named_param('a')
    assert factory_b.has_typed_param(A)

    assert factory_b.check_signature(Callable[[A], B])
    assert factory_b.check_signature(Callable[[A], object])

    res_b = factory_b.create(a=res_a)
    assert isinstance(res_b, B)
    assert res_b.a is res_a

    assert factory_b.add_typed_arg(A, res_a)

    assert factory_b.check_signature(Callable[[A], B])
    assert factory_b.check_signature(Callable[[A], object])
    assert factory_b.check_signature(Callable[[], B])
    assert factory_b.check_signature(Callable[[], object])

    res_b = factory_b.create()
    assert isinstance(res_b, B)
    assert res_b.a is res_a

    factory_c = into_factory(C)

    assert not factory_c.scoped
    assert factory_c.cacheable
    assert not factory_c.is_async

    assert factory_c.return_type is C
    assert factory_c.params_count == 1
    assert factory_c.args_count == 0
    assert factory_c.actual_params == {'conf': str}

    assert factory_c.has_named_param('conf')
    assert factory_c.has_typed_param(str)

    assert factory_c.check_signature(Callable[[str], C])
    assert not factory_c.check_signature(Callable[[], C])

    assert factory_c.add_named_arg('conf', 'test')
    assert factory_c.check_signature(Callable[[], C])

    res_c = factory_c.create()
    assert isinstance(res_c, C)
    assert res_c.conf == 'test'

    factory_d = into_factory(D)

    assert not factory_d.scoped
    assert factory_d.cacheable
    assert not factory_d.is_async

    assert factory_d.return_type is D
    assert factory_d.params_count == 3
    assert factory_d.args_count == 0
    assert factory_d.actual_params == {'a': A, 'b': Maybe[B], 'c': C | None}

    assert factory_d.has_named_param('a')
    assert factory_d.has_named_param('b')
    assert factory_d.has_named_param('c')
    assert factory_d.has_typed_param(A)
    assert factory_d.has_typed_param(B)
    assert factory_d.has_typed_param(C)

    assert factory_d.check_signature(Callable[[A, B, C], D])
    assert factory_d.check_signature(Callable[[A, B], D])
    assert factory_d.check_signature(Callable[[A, C], D])
    assert factory_d.check_signature(Callable[[A], D])
    assert factory_d.check_signature(Callable[[A, Maybe[B], C], D])
    assert factory_d.check_signature(Callable[[A, Maybe[B], C | None], D])
    assert factory_d.check_signature(Callable[[A, B, C | None], D])
    assert not factory_d.check_signature(Callable[[B, C | None], D])

    assert factory_d.add_named_arg('a', res_a)

    assert factory_d.check_signature(Callable[[B, C], D])
    assert factory_d.check_signature(Callable[[B], D])
    assert factory_d.check_signature(Callable[[C], D])
    assert factory_d.check_signature(Callable[[], D])
    assert factory_d.check_signature(Callable[[Maybe[B], C], D])
    assert factory_d.check_signature(Callable[[Maybe[B], C | None], D])
    assert factory_d.check_signature(Callable[[B, C | None], D])

    res_d = factory_d.create()
    assert isinstance(res_d, D)
    assert res_d.a is res_a
    assert res_d.b.is_nothing()
    assert res_d.c is None

    assert factory_d.add_typed_arg(B, res_b)

    assert factory_d.check_signature(Callable[[B, C], D])
    assert factory_d.check_signature(Callable[[B], D])
    assert factory_d.check_signature(Callable[[C], D])
    assert factory_d.check_signature(Callable[[], D])
    assert factory_d.check_signature(Callable[[Maybe[B], C], D])
    assert factory_d.check_signature(Callable[[Maybe[B], C | None], D])
    assert factory_d.check_signature(Callable[[B, C | None], D])

    res_d = factory_d.create()
    assert isinstance(res_d, D)
    assert res_d.a is res_a
    assert res_d.b.unwrap() is res_b
    assert res_d.c is None

    assert factory_d.add_typed_arg(C, res_c)

    assert factory_d.check_signature(Callable[[B, C], D])
    assert factory_d.check_signature(Callable[[B], D])
    assert factory_d.check_signature(Callable[[C], D])
    assert factory_d.check_signature(Callable[[], D])
    assert factory_d.check_signature(Callable[[Maybe[B], C], D])
    assert factory_d.check_signature(Callable[[Maybe[B], C | None], D])
    assert factory_d.check_signature(Callable[[B, C | None], D])

    res_d = factory_d.create()
    assert isinstance(res_d, D)
    assert res_d.a is res_a
    assert res_d.b.unwrap() is res_b
    assert res_d.c is res_c

    res_d = factory_d.create(res_b, res_c)
    assert isinstance(res_d, D)
    assert res_d.a is res_a
    assert res_d.b.unwrap() is res_b
    assert res_d.c is res_c

    res_d = factory_d.create(b=res_b, c=res_c)
    assert isinstance(res_d, D)
    assert res_d.a is res_a
    assert res_d.b.unwrap() is res_b
    assert res_d.c is res_c


@pytest.mark.parametrize(
    'param', ((sync_simple_fn, False), (async_simple_fn, True)), ids=('Sync', 'Async')
)
@pytest.mark.asyncio
async def test_simple_callable(param) -> None:  # noqa: ANN001
    orig_fn = param[0]
    is_async = param[1]

    factory = into_factory(orig_fn)

    assert not factory.scoped
    assert not factory.cacheable
    assert is_async == factory.is_async

    assert factory.return_type is int
    assert factory.params_count == 0
    assert factory.args_count == 0
    assert factory.actual_params == {}

    assert factory.check_signature(Callable[[], int])
    assert factory.check_signature(Callable[[], object])

    if not is_async:
        res = factory.create()
    else:
        res = await factory.create()

    assert isinstance(res, int)
    assert res == 123


@pytest.mark.parametrize(
    'param',
    (
        (sync_complex_fn, False, False),
        (async_complex_fn, True, False),
        (sync_ctx_fn, False, True),
        (async_ctx_fn, True, True),
    ),
    ids=('Sync', 'Async', 'SyncCtx', 'AsyncCtx'),
)
@pytest.mark.asyncio
async def test_complex_callable(param) -> None:  # noqa: ANN001
    orig_fn = param[0]
    is_async = param[1]
    is_scoped = param[2]

    factory = into_factory(orig_fn)

    assert not factory.cacheable
    assert is_scoped == factory.scoped
    assert is_async == factory.is_async

    assert factory.return_type is Decimal
    assert factory.params_count == 2
    assert factory.args_count == 0
    assert factory.actual_params == {'a1': int, 'a2': str}

    assert not factory.has_named_param('a0')
    assert factory.has_named_param('a1')
    assert factory.has_named_param('a2')

    assert not factory.has_typed_param(bool)
    assert factory.has_typed_param(int)
    assert factory.has_typed_param(str)

    assert factory.check_signature(Callable[[int, str], Decimal])

    if not is_async:
        res = factory.create(1, '2')
    else:
        res = await factory.create(1, '2')

    assert isinstance(res, Decimal)
    assert res == Decimal('3')

    if not is_async:
        factory.destroy(res)
    else:
        await factory.destroy(res)


@pytest.mark.parametrize(
    'param',
    (
        (sync_complex_fn, False, False),
        (async_complex_fn, True, False),
        (sync_ctx_fn, False, True),
        (async_ctx_fn, True, True),
    ),
    ids=('Sync', 'Async', 'SyncCtx', 'AsyncCtx'),
)
@pytest.mark.asyncio
async def test_complex_override(param) -> None:  # noqa: ANN001
    orig_fn = param[0]
    is_async = param[1]

    factory = into_factory(orig_fn)

    assert not factory.check_signature(Callable[[], int])
    assert not factory.check_signature(Callable[[], Decimal])
    assert not factory.check_signature(Callable[[int], Decimal])
    assert not factory.check_signature(Callable[[str], Decimal])
    assert factory.check_signature(Callable[[int, str], Decimal])

    assert factory.args_count == 0
    assert factory.actual_params == {'a1': int, 'a2': str}

    assert not factory.add_named_arg('a0', 123)
    assert not factory.add_typed_arg(bool, True)

    assert factory.add_named_arg('a1', 123)
    assert factory.args_count == 1
    assert factory.actual_params == {'a2': str}

    assert not factory.check_signature(Callable[[], Decimal])
    assert not factory.check_signature(Callable[[int], Decimal])
    assert factory.check_signature(Callable[[str], Decimal])
    assert factory.check_signature(Callable[[int, str], Decimal])

    assert factory.add_typed_arg(str, '456')
    assert factory.args_count == 2
    assert factory.actual_params == {}

    assert factory.check_signature(Callable[[], Decimal])
    assert factory.check_signature(Callable[[int], Decimal])
    assert factory.check_signature(Callable[[str], Decimal])
    assert factory.check_signature(Callable[[int, str], Decimal])

    if not is_async:
        res = factory.create()
    else:
        res = await factory.create()

    assert isinstance(res, Decimal)
    assert res == Decimal('579')

    if not is_async:
        factory.destroy(res)
    else:
        await factory.destroy(res)

    if not is_async:
        res = factory.create(a2='789')
    else:
        res = await factory.create(a2='789')

    assert isinstance(res, Decimal)
    assert res == Decimal('912')

    if not is_async:
        factory.destroy(res)
    else:
        await factory.destroy(res)

    assert factory.remove_typed_arg(int)
    assert factory.args_count == 1
    assert factory.actual_params == {'a1': int}

    assert not factory.check_signature(Callable[[], Decimal])
    assert factory.check_signature(Callable[[int], Decimal])
    assert not factory.check_signature(Callable[[str], Decimal])
    assert factory.check_signature(Callable[[int, str], Decimal])

    assert factory.remove_named_arg('a2')
    assert factory.args_count == 0
    assert factory.actual_params == {'a1': int, 'a2': str}

    assert not factory.check_signature(Callable[[], Decimal])
    assert not factory.check_signature(Callable[[int], Decimal])
    assert not factory.check_signature(Callable[[str], Decimal])
    assert factory.check_signature(Callable[[int, str], Decimal])

    assert not factory.remove_named_arg('a0')
    assert not factory.remove_typed_arg(bool)


def test_sync_scope(faker: Faker) -> None:
    gen_mock = Mock()

    def gen() -> Generator[str, str, None]:
        gen_mock()
        yield faker.pystr()
        gen_mock()

    factory = into_factory(gen)
    assert factory.scoped
    assert factory.return_type is str
    assert not factory.is_async

    factory.destroy(factory.create())

    assert gen_mock.call_count == 2


@pytest.mark.asyncio
async def test_async_scope(faker: Faker) -> None:
    gen_mock = Mock()

    async def gen() -> AsyncGenerator[str, str, None]:
        gen_mock()
        yield faker.pystr()
        gen_mock()

    factory = into_factory(gen)
    assert factory.scoped
    assert factory.return_type is str
    assert factory.is_async

    await factory.destroy(await factory.create())

    assert gen_mock.call_count == 2



import weakref
from dataclasses import dataclass
from typing import Any, override

from greyhorse.app.abc.collections.collectors import MutCollector
from greyhorse.app.abc.resources.common import ResourceEventKind
from greyhorse.app.abc.resources.operators import (
    OperatorData,
    OperatorEvent,
    OperatorListenerFn,
    OperatorSubscription,
)
from greyhorse.app.abc.resources.providers import (
    ProviderData,
    ProviderEvent,
    ProviderListenerFn,
    ProviderSubscription,
)

from ..collections.registries import MutDictRegistry
from ..runtime.invoke import invoke_sync


@dataclass(slots=True, frozen=True, kw_only=True)
class _CovariantListenersFilter[T]:
    kind: ResourceEventKind
    key: type[T]

    def __call__(self, t: type[T], metadata: dict[str, Any]) -> bool:
        return t is Any or (
            issubclass(self.key, t) and metadata['kind'] & self.kind != ResourceEventKind(0)
        )


@dataclass(slots=True, frozen=True, kw_only=True)
class _ContravariantListenersFilter[T]:
    kind: ResourceEventKind
    key: type[T]

    def __call__(self, t: type[T], metadata: dict[str, Any]) -> bool:
        return t is Any or (
            issubclass(t, self.key) and metadata['kind'] & self.kind != ResourceEventKind(0)
        )


def _is_bound_method(cb: Any) -> bool:
    return (
        hasattr(cb, '__self__')
        and hasattr(cb, '__func__')
        and cb.__self__ is not None
        and not isinstance(cb.__self__, type)
        and getattr(cb.__func__, '__self__', None) is None
    )


class OperatorRegistry(MutCollector[type, OperatorData], OperatorSubscription):
    __slots__ = ('_metadata', '_op_listeners')

    def __init__(self) -> None:
        super().__init__()
        self._metadata = MutDictRegistry[type, OperatorData](allow_many=True)
        self._op_listeners = MutDictRegistry[type, OperatorListenerFn](allow_many=True)

    @override
    def add[U](self, target_type: type[U], instance: OperatorData, /, **metadata: Any) -> bool:
        filter = _CovariantListenersFilter(kind=ResourceEventKind.SETUP, key=target_type)

        event = OperatorEvent(kind=ResourceEventKind.SETUP, type=target_type, data=instance)
        self._notify(event, filter)
        return self._metadata.add(target_type, instance, **metadata)

    @override
    def remove[U](self, target_type: type[U], instance: OperatorData | None = None) -> bool:
        filter = _CovariantListenersFilter(kind=ResourceEventKind.TEARDOWN, key=target_type)

        if instance is None:
            for data, _md in self._metadata.list_with_metadata(target_type):
                event = OperatorEvent(
                    kind=ResourceEventKind.TEARDOWN, type=target_type, data=data
                )
                self._notify(event, filter)
        else:
            event = OperatorEvent(
                kind=ResourceEventKind.TEARDOWN, type=target_type, data=instance
            )
            self._notify(event, filter)

        return self._metadata.remove(target_type, instance)

    @override
    def add_operator_listener[T](
        self,
        key: type[T],
        fn: OperatorListenerFn,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool:
        if _is_bound_method(fn):
            fn = weakref.WeakMethod(fn)
            md = {'weak': True}
        else:
            md = {'weak': False}
        return self._op_listeners.add(key, fn, kind=kind, **md)

    @override
    def remove_operator_listener[T](
        self,
        key: type[T],
        fn: OperatorListenerFn | None = None,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool:
        filter = _CovariantListenersFilter(kind=kind, key=key)
        to_remove = []

        for t, orig_fn, metadata in self._op_listeners.filter(filter):
            if metadata['weak'] and (orig_fn := orig_fn()) is None:
                to_remove.append((t, orig_fn, metadata['kind']))
                continue
            if fn is None or fn == orig_fn:
                to_remove.append((t, orig_fn, metadata['kind']))

        res = False

        for orig_key, orig_fn, orig_kind in to_remove:
            new_kind = orig_kind & ~kind
            res |= self._op_listeners.remove(orig_key, orig_fn)
            if new_kind != ResourceEventKind(0):
                self._op_listeners.add(orig_key, orig_fn, kind=new_kind)

        return res

    def _notify(self, event: OperatorEvent, filter: _CovariantListenersFilter) -> None:
        callbacks = []
        to_remove = []

        for t, listener, metadata in self._op_listeners.filter(filter):
            if metadata['weak']:
                if listener._alive:  # noqa: SLF001
                    listener = listener()
                else:
                    to_remove.append((t, listener, metadata['kind']))
                    continue
            callbacks.append(listener)

        for orig_key, orig_fn, _orig_kind in to_remove:
            self._op_listeners.remove(orig_key, orig_fn)

        for listener in callbacks:
            invoke_sync(listener, event)


class ProviderRegistry(MutCollector[type, ProviderData], ProviderSubscription):
    __slots__ = ('_metadata', '_prov_listeners')

    def __init__(self) -> None:
        super().__init__()
        self._metadata = MutDictRegistry[type, ProviderData](allow_many=True)
        self._prov_listeners = MutDictRegistry[type, ProviderListenerFn](allow_many=True)

    @override
    def add[U](self, target_type: type[U], instance: ProviderData, /, **metadata: Any) -> bool:
        filter = _CovariantListenersFilter(kind=ResourceEventKind.SETUP, key=target_type)

        event = ProviderEvent(kind=ResourceEventKind.SETUP, type=target_type, data=instance)
        self._notify(event, filter)
        return self._metadata.add(target_type, instance, **metadata)

    @override
    def remove[U](self, target_type: type[U], instance: ProviderData | None = None) -> bool:
        filter = _CovariantListenersFilter(kind=ResourceEventKind.TEARDOWN, key=target_type)

        if instance is None:
            for data, _md in self._metadata.list_with_metadata(target_type):
                event = ProviderEvent(
                    kind=ResourceEventKind.TEARDOWN, type=target_type, data=data
                )
                self._notify(event, filter)
        else:
            event = ProviderEvent(
                kind=ResourceEventKind.TEARDOWN, type=target_type, data=instance
            )
            self._notify(event, filter)

        return self._metadata.remove(target_type, instance)

    @override
    def add_provider_listener[T](
        self,
        key: type[T],
        fn: ProviderListenerFn,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool:
        if _is_bound_method(fn):
            fn = weakref.WeakMethod(fn)
            md = {'weak': True}
        else:
            md = {'weak': False}
        return self._prov_listeners.add(key, fn, kind=kind, **md)

    @override
    def remove_provider_listener[T](
        self,
        key: type[T],
        fn: ProviderListenerFn | None = None,
        kind: ResourceEventKind = ResourceEventKind.ALL,
    ) -> bool:
        filter = _CovariantListenersFilter(kind=kind, key=key)
        to_remove = []

        for t, orig_fn, metadata in self._prov_listeners.filter(filter):
            if metadata['weak'] and (orig_fn := orig_fn()) is None:
                to_remove.append((t, orig_fn, metadata['kind']))
                continue
            if fn is None or fn == orig_fn:
                to_remove.append((t, orig_fn, metadata['kind']))

        res = False

        for orig_key, orig_fn, orig_kind in to_remove:
            new_kind = orig_kind & ~kind
            res |= self._prov_listeners.remove(orig_key, orig_fn)
            if new_kind != ResourceEventKind(0):
                self._prov_listeners.add(orig_key, orig_fn, kind=new_kind)

        return res

    def _notify(self, event: ProviderEvent, filter: _CovariantListenersFilter) -> None:
        callbacks = []
        to_remove = []

        for t, listener, metadata in self._prov_listeners.filter(filter):
            if metadata['weak']:
                if listener._alive:  # noqa: SLF001
                    listener = listener()
                else:
                    to_remove.append((t, listener, metadata['kind']))
                    continue
            callbacks.append(listener)

        for orig_key, orig_fn, _orig_kind in to_remove:
            self._prov_listeners.remove(orig_key, orig_fn)

        for listener in callbacks:
            invoke_sync(listener, event)



from abc import ABC, abstractmethod
from collections.abc import Awaitable

from greyhorse.result import Result

from ..collections.selectors import ListSelector
from .borrow import BorrowError, BorrowMutError
from .common import ResourceStatus, TransitionStatus


class BaseResource(ABC):
    @property
    @abstractmethod
    def status(self) -> ResourceStatus: ...

    @property
    @abstractmethod
    def transition_status(self) -> TransitionStatus: ...


class ResourceState[S](BaseResource, ABC):
    @property
    @abstractmethod
    def shared_type(self) -> type[S]: ...

    @property
    @abstractmethod
    def next_status(self) -> ResourceStatus | None: ...

    @abstractmethod
    def borrow(self) -> Result[S, BorrowError] | Awaitable[Result[S, BorrowError]]: ...

    @abstractmethod
    def reclaim(self) -> None | Awaitable[None]: ...


class MutResourceState[S, M](BaseResource, ABC):
    @property
    @abstractmethod
    def mut_type(self) -> type[M]: ...

    @property
    @abstractmethod
    def next_status(self) -> ResourceStatus | None: ...

    @abstractmethod
    def acquire(self) -> Result[M, BorrowMutError] | Awaitable[Result[M, BorrowMutError]]: ...

    @abstractmethod
    def release(self) -> None | Awaitable[None]: ...


ResourceSelector = ListSelector[type, ResourceState]
MutResourceSelector = ListSelector[type, MutResourceState]



import inspect
import types
from collections.abc import Callable, Iterable, Mapping
from dataclasses import dataclass, field
from typing import Any

from greyhorse.app.abc.resources.common import ResourceEventKind
from greyhorse.app.abc.resources.operators import OperatorListenerFn, OperatorSubscription
from greyhorse.app.abc.resources.reconciler import ReconcilerCollector


@dataclass(slots=True, frozen=True, kw_only=True)
class _OperatorListenerData:
    target: OperatorListenerFn
    for_type: type | types.UnionType
    kind: ResourceEventKind
    metadata: Mapping[str, Any] = field(default_factory=dict)


def _is_operator_listener_member(member: types.FunctionType) -> bool:
    return hasattr(member, '__op_list_data__')


def operator_listener(
    for_type: type | types.UnionType,
    kind: ResourceEventKind = ResourceEventKind.ALL,
    **metadata: Any,
) -> Callable[[OperatorListenerFn], OperatorListenerFn]:
    def decorator(func: OperatorListenerFn) -> OperatorListenerFn:
        func.__op_list_data__ = _OperatorListenerData(
            target=func, for_type=for_type, kind=kind, metadata=metadata
        )
        return func

    return decorator


class Controller:
    __slots__ = ('__op_listeners',)

    def __init__(self) -> None:
        self.__op_listeners: list[_OperatorListenerData] = []
        self._init_members()

    def setup(
        self, op_subscription: OperatorSubscription, task_collector: ReconcilerCollector
    ) -> None:
        for data in self.__op_listeners:
            target_types = self._get_target_types(data)
            for target_type in target_types:
                op_subscription.add_operator_listener(target_type, data.target, data.kind)

    def teardown(self, op_subscription: OperatorSubscription) -> None:
        for data in reversed(self.__op_listeners):
            target_types = self._get_target_types(data)
            for target_type in target_types:
                op_subscription.remove_operator_listener(target_type, data.target, data.kind)

    def _init_members(self) -> None:
        for _, member in inspect.getmembers(self):
            if _is_operator_listener_member(member):
                data = member.__op_list_data__
                # data = replace(data, target=into_factory(member))
                self.__op_listeners.append(data)

    def _get_target_types(self, data: _OperatorListenerData) -> Iterable[type]:
        if isinstance(data.for_type, types.UnionType):
            target_types = data.for_type.__args__
        else:
            target_types = [data.for_type]
        return target_types



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass, field
from enum import IntEnum
from types import TracebackType
from typing import Self

from greyhorse.factory import Factory
from greyhorse.maybe import Maybe


@dataclass(slots=True, frozen=True, kw_only=True)
class ResolvedData[T]:
    type: type[T]
    factory: Factory[T]
    deps: dict[str, type] = field(default_factory=dict)


@dataclass(slots=True, frozen=True, kw_only=True)
class ResolveResult:
    resolved: dict[type, ResolvedData] = field(default_factory=dict)
    unresolved: list[type] = field(default_factory=list)


class Resolver(ABC):
    @abstractmethod
    def can_resolve[T](self, key: type[T]) -> bool: ...

    @abstractmethod
    def resolve_factories[T](
        self, key: type[T], from_scope: IntEnum | None = None
    ) -> ResolveResult: ...

    @abstractmethod
    def resolve_value[T](
        self,
        resolved: Mapping[type, ResolvedData],
        key: type[T],
        from_scope: IntEnum | None = None,
    ) -> Maybe[T]: ...

    @abstractmethod
    def resolve[T](self, key: type[T], from_scope: IntEnum | None = None) -> Maybe[T]: ...

    @abstractmethod
    def __enter__(self) -> Self: ...

    @abstractmethod
    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None: ...



from __future__ import annotations

import asyncio
import enum
import threading
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable, Collection
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Optional

from greyhorse.enum import Enum, Struct, Unit
from greyhorse.utils.types import TypeWrapper

from .common import Condition, ResourceStatus, TransitionStatus


class ReconcilerRunPhase(Enum):
    Setup = Unit()
    Teardown = Unit()
    Periodic = Struct(interval=timedelta)
    Permanent = Struct(async_=bool)


class RestartPolicy(Enum):
    Always = Unit()
    Never = Unit()
    OnFailure = Unit()


type ReconcilerTaskFn[T] = Callable[[TaskHandler], T | Awaitable[T]]


class ReconcilerCollector(ABC):
    @abstractmethod
    def add_task[T](
        self,
        key: type[T],
        function: ReconcilerTaskFn[T],
        phase: ReconcilerRunPhase = ReconcilerRunPhase.Setup,
        restart_policy: RestartPolicy = RestartPolicy.Never,
        conditions: list[str] | None = None,
    ) -> bool: ...

    @abstractmethod
    def remove_task[T](
        self,
        key: type[T],
        function: ReconcilerTaskFn[T],
        phase: ReconcilerRunPhase = ReconcilerRunPhase.Setup,
    ) -> bool: ...


class TaskStatus(Enum):
    Pending = Unit()
    Scheduled = Struct(at=datetime)
    Running = Unit()
    Canceled = Struct(reason=str)
    Completed = Struct(result=Any)
    Failed = Struct(message=str, exc=Optional[Exception])  # noqa: UP045


class ReconcilerEventType(enum.IntEnum):
    INFO = enum.auto()
    WARN = enum.auto()
    OK = enum.auto()
    ERR = enum.auto()


@dataclass(slots=True, frozen=True, kw_only=True)
class ReconcilerEvent:
    type: ReconcilerEventType
    phase: ReconcilerRunPhase
    regarding: str | None = None
    action: str | None = None
    reason: str | None = None
    message: str
    reporter: str
    timestamp: datetime = field(default_factory=datetime.now)


class Task(ABC):
    @property
    @abstractmethod
    def status(self) -> TaskStatus: ...

    @property
    @abstractmethod
    def restart_policy(self) -> RestartPolicy: ...

    @property
    @abstractmethod
    def phase(self) -> ReconcilerRunPhase: ...

    @abstractmethod
    def events(self) -> Collection[ReconcilerEvent]: ...


class TaskHandler(ABC):
    @abstractmethod
    def add_event(self, event: ReconcilerEvent) -> None: ...

    @abstractmethod
    def set_condition(self, condition: Condition) -> bool: ...

    @abstractmethod
    def cancel(self, reason: str | None = None) -> None: ...

    @abstractmethod
    def complete[T](self, value: T | None = None) -> None: ...

    @abstractmethod
    def failure(self, message: str, exception: Exception | None = None) -> None: ...


class ReconcilerStatus(Enum):
    Dead = Unit()
    Alive = Struct(status=TransitionStatus)
    Succeeded = Struct(status=TransitionStatus)
    Failed = Struct(status=TransitionStatus, message=str)


class ReconcilerWaiter(Enum):
    Sync = Struct(value=threading.Event)
    Async = Struct(value=asyncio.Event)


class Reconciler[S](TypeWrapper[S], ABC):
    @property
    @abstractmethod
    def status(self) -> ReconcilerStatus: ...

    @property
    @abstractmethod
    def waiter(self) -> ReconcilerWaiter | None: ...

    @abstractmethod
    def set_resource_status(self, status: ResourceStatus) -> None: ...

    @abstractmethod
    def set_condition(self, condition: Condition) -> None: ...

    @abstractmethod
    def reset_condition(self, name: str) -> bool: ...

    @abstractmethod
    def events(self) -> Collection[ReconcilerEvent]: ...

    @abstractmethod
    def setup(self, task_collector: ReconcilerCollector) -> None | Awaitable[None]: ...

    @abstractmethod
    def teardown(self) -> None | Awaitable[None]: ...



import asyncio
import inspect
import threading
import types
from collections.abc import Callable, Iterable, Mapping
from dataclasses import dataclass, field
from typing import Any

from greyhorse.app.abc.resources.common import ResourceEventKind
from greyhorse.app.abc.resources.providers import ProviderListenerFn, ProviderSubscription
from greyhorse.app.abc.resources.reconciler import ReconcilerCollector
from greyhorse.enum import Enum, Struct, Unit


@dataclass(slots=True, frozen=True, kw_only=True)
class _ProviderListenerData:
    target: ProviderListenerFn
    for_type: type | types.UnionType
    kind: ResourceEventKind
    metadata: Mapping[str, Any] = field(default_factory=dict)


def _is_provider_listener_member(member: types.FunctionType) -> bool:
    return hasattr(member, '__prov_list_data__')


def provider_listener(
    for_type: type | types.UnionType,
    kind: ResourceEventKind = ResourceEventKind.ALL,
    **metadata: Any,
) -> Callable[[ProviderListenerFn], ProviderListenerFn]:
    def decorator(func: ProviderListenerFn) -> ProviderListenerFn:
        func.__prov_list_data__ = _ProviderListenerData(
            target=func, for_type=for_type, kind=kind, metadata=metadata
        )
        return func

    return decorator


class ServiceState(Enum):
    Idle = Unit()
    Initialization = Unit()
    Finalization = Unit()
    Initialized = Unit()
    Finalized = Unit()
    Ready = Struct(started=bool)
    Unready = Struct(started=bool, message=str)


class ServiceWaiter(Enum):
    Sync = Struct(value=threading.Event)
    Async = Struct(value=asyncio.Event)


class Service:
    __slots__ = ('__prov_listeners', '__state', '__waiter')

    def __init__(self, waiter: ServiceWaiter) -> None:
        self.__state = ServiceState.Idle
        self.__waiter = waiter
        self.__prov_listeners: list[_ProviderListenerData] = []
        self._init_members()

    @property
    def state(self) -> ServiceState:
        return self.__state

    @property
    def waiter(self) -> ServiceWaiter:
        return self.__waiter

    def setup(
        self, prov_subscription: ProviderSubscription, task_collector: ReconcilerCollector
    ) -> None:
        for data in self.__prov_listeners:
            target_types = self._get_target_types(data)
            for target_type in target_types:
                prov_subscription.add_provider_listener(target_type, data.target, data.kind)

    def teardown(self, prov_subscription: ProviderSubscription) -> None:
        for data in reversed(self.__prov_listeners):
            target_types = self._get_target_types(data)
            for target_type in target_types:
                prov_subscription.remove_provider_listener(target_type, data.target, data.kind)

    def _init_members(self) -> None:
        for _, member in inspect.getmembers(self):
            if _is_provider_listener_member(member):
                data = member.__prov_list_data__
                # data = replace(data, target=into_factory(member))
                self.__prov_listeners.append(data)

    def _get_target_types(self, data: _ProviderListenerData) -> Iterable[type]:
        if isinstance(data.for_type, types.UnionType):
            target_types = data.for_type.__args__
        else:
            target_types = [data.for_type]
        return target_types

    def _set_state(self, state: ServiceState) -> None:
        match state:
            case ServiceState.Ready(started) | ServiceState.Unready(started):
                if started:
                    self.__waiter.value.clear()
                else:
                    self.__waiter.value.set()
            case _:
                pass

        self.__state = state



from typing import Any, ForwardRef

from pydantic import BaseModel, Field, PrivateAttr

from ..fragment import Fragment
from ..runtime.invoke import caller_path
from .ctrl import CtrlConf
from .svc import SvcConf


class ComponentConf(BaseModel, frozen=True):
    enabled: bool = Field(default=True)
    fragments: list[type[Fragment]] = Field(default_factory=list)
    exports: list[type] = Field(default_factory=list)


class ModuleComponentConf(ComponentConf, frozen=True):
    name: str
    path: str = Field(frozen=True)
    args: dict[str, Any] = Field(default_factory=dict, frozen=True)
    _init_path: list[str] = PrivateAttr(default_factory=lambda: caller_path(5))
    _conf: ForwardRef('ModuleConf') | None = PrivateAttr(default=None)


class ModuleConf(BaseModel, frozen=True):
    enabled: bool = Field(default=True)

    controllers: list[CtrlConf] = Field(default_factory=list)
    services: list[SvcConf] = Field(default_factory=list)

    fragments: list[type[Fragment]] = Field(default_factory=list)
    components: dict[str, ComponentConf] = Field(default_factory=dict)



import asyncio
from collections.abc import Awaitable, Callable
from functools import partial, wraps

from greyhorse.app.private.runtime.runtime import Runtime, get_runtime
from greyhorse.utils.types import is_awaitable


def wrap_sync[T, **P](
    func: Callable[P, Awaitable[T]], /, *args: P.args, **kwargs: P.kwargs
) -> T:
    runtime = get_runtime()
    runtime.start()
    try:
        return func(*args, **kwargs)
    finally:
        runtime.stop()
        Runtime._instance = None  # noqa: SLF001


async def wrap_async[T, **P](
    func: Callable[P, Awaitable[T]], /, *args: P.args, **kwargs: P.kwargs
) -> T:
    runtime = get_runtime()
    await runtime.start()
    try:
        return await func(*args, **kwargs)
    finally:
        await runtime.stop()
        Runtime._instance = None  # noqa: SLF001


def run[T, **P](func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs) -> T:
    if is_awaitable(func):
        asyncio.set_event_loop(asyncio.new_event_loop())
        return asyncio.run(wrap_async(func, *args, **kwargs))
    return wrap_sync(func, *args, **kwargs)


def main[T, **P](func: Callable[P, T] | None = None) -> Callable[P, T]:
    @wraps(func)
    def decorator[T, **P](func: Callable[P, T]) -> T:
        return partial(run, func)

    if func is None:
        return decorator

    return decorator(func)



import asyncio
import sys
from asyncio import iscoroutine
from collections.abc import Callable
from typing import override

import greenback
from greenlet import getcurrent, greenlet

from greyhorse.app.abc.runtime.runtime import Runtime
from greyhorse.utils.types import is_awaitable


class _Greenlet(greenlet):
    __slots__ = ('main_context',)

    def __init__[T, **P](self, func: Callable[P, T], async_context: greenlet) -> None:
        greenlet.__init__(self, func, async_context)
        self.main_context = async_context


class SyncRuntime(Runtime):
    __slots__ = ('_counter', '_executor')

    def __init__(self) -> None:
        super().__init__(asyncio.new_event_loop())
        self._executor = _Greenlet(self._loop.run_forever, getcurrent())
        self._counter = 0

    @property
    @override
    def active(self) -> bool:
        return not self._executor.dead

    async def _loop_fn(self) -> None:
        result = self._executor.main_context.switch()

        while result != ():
            try:
                value = await result
            except BaseException:
                result = self._executor.main_context.throw(*sys.exc_info())
            else:
                result = self._executor.main_context.switch(value)

    @override
    def start(self) -> None:
        if self._counter == 0:
            self._loop.create_task(self._loop_fn())
            self._executor.switch()
        self._counter += 1

    @override
    def stop(self) -> None:
        if self._counter == 1:
            if not self._loop.is_closed():
                self._loop.call_soon_threadsafe(self._loop.stop)
            self._executor.switch()
            if not self._loop.is_closed():
                self._loop.close()

        self._counter = max(self._counter - 1, 0)

    @override
    def invoke_sync[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        if is_awaitable(func):
            if not self.active:
                return asyncio.run(func(*args, **kwargs))
            return self._executor.switch(func(*args, **kwargs))

        return func(*args, **kwargs)

    @override
    async def invoke_async[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        if is_awaitable(func):
            if iscoroutine(func):
                return await func  # type: ignore
            return await func(*args, **kwargs)

        context = _Greenlet(func, getcurrent())
        try:
            result = context.switch(*args, **kwargs)

            while not context.dead:
                try:
                    value = await result

                except BaseException:
                    result = context.throw(*sys.exc_info())
                else:
                    result = context.switch(value)
        finally:
            # clean up to avoid cycle resolution by gc
            del context.main_context

        return result


class AsyncRuntime(Runtime):
    __slots__ = ('_counter',)

    def __init__(self) -> None:
        super().__init__(asyncio.get_running_loop())
        self._counter = 0

    @property
    @override
    def active(self) -> bool:
        return self._loop.is_running()

    @override
    async def start(self) -> None:
        if self._counter == 0:
            await greenback.ensure_portal()
        self._counter += 1

    @override
    async def stop(self) -> None:
        self._counter = max(self._counter - 1, 0)

    @override
    def invoke_sync[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        if is_awaitable(func):
            return greenback.await_(func(*args, **kwargs))
        return func(*args, **kwargs)

    @override
    async def invoke_async[T, **P](
        self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
    ) -> T:
        if is_awaitable(func):
            if iscoroutine(func):
                return await func  # type: ignore
            return await greenback.with_portal_run(func, *args, **kwargs)
        return await greenback.with_portal_run_sync(func, *args, **kwargs)


def get_runtime() -> Runtime:
    if Runtime._instance is not None:  # noqa: SLF001
        return Runtime._instance  # noqa: SLF001

    try:
        asyncio.get_running_loop()
        instance = AsyncRuntime()
    except RuntimeError:
        instance = SyncRuntime()

    Runtime._instance = instance  # noqa: SLF001
    return instance



from __future__ import annotations

from collections.abc import Mapping
from enum import IntEnum
from types import TracebackType
from typing import Any, override

from greyhorse.app.abc.collections.selectors import Selector
from greyhorse.app.abc.functional.context import OperationContext
from greyhorse.app.abc.functional.registries import FactoryRegistry
from greyhorse.app.abc.functional.resolver import Resolver
from greyhorse.maybe import Maybe

from .bucket import Bucket
from .resolver import BucketResolver


class _ResolverSelector(Selector[type, Any]):
    __slots__ = ('_resolver', '_scope')

    def __init__(
        self, bucket: Bucket, registry: FactoryRegistry, filters: Mapping[str, Any]
    ) -> None:
        self._resolver = BucketResolver(registry, bucket, **filters)
        self._scope = bucket.scope

    @property
    def resolver(self) -> BucketResolver:
        return self._resolver

    def __repr__(self) -> str:
        if self._scope is not None:
            return f'ResolverSelector <{self._scope.name.lower()}>'
        return 'ResolverSelector'

    @override
    def has[T](self, key: type[T]) -> bool:
        return self._resolver.can_resolve(key)

    @override
    def get[T](self, key: type[T]) -> Maybe[T]:
        return self._resolver.resolve(key, self._scope)

    @override
    def get_with_metadata[T](self, key: type[T]) -> Maybe[tuple[T, Mapping[str, Any]]]:
        return self._resolver.resolve(key, self._scope).map(lambda x: (x, {}))


class BucketOperationContext(OperationContext):
    __slots__ = ('_bucket', '_filters', '_registry', '_scope_class')

    def __init__[S: IntEnum](
        self,
        registry: FactoryRegistry,
        scope_class: type[S] | None = None,
        bucket: Bucket | None = None,
        /,
        **filters: Any,
    ) -> None:
        super().__init__(_ResolverSelector)
        self._registry = registry
        self._scope_class = scope_class
        self._filters = filters

        if bucket:
            self._bucket = bucket
        else:
            self._bucket = Bucket(scope=self._scope_class(0)) if self._scope_class else Bucket()

    @override
    @property
    def scope[S: IntEnum](self) -> S | None:
        return self._bucket.scope

    @override
    def context[S: IntEnum](self, scope: S | None = None) -> OperationContext:
        bucket = self._bucket

        if self._scope_class and scope is not None:
            if scope < bucket.scope:
                while scope < bucket.scope and bucket.prev is not None:
                    bucket = bucket.prev
            elif scope > self._bucket.scope:
                next_scope = self._scope_class(self._bucket.scope + 1)
                bucket = self._construct_chain(bucket, next_scope, scope)

        return BucketOperationContext(
            self._registry, self._scope_class, bucket, **self._filters
        )

    @override
    def context_resolver[S: IntEnum](self, scope: S | None = None) -> Resolver:
        bucket = self._bucket

        if self._scope_class and scope is not None:
            if scope < bucket.scope:
                while scope < bucket.scope and bucket.prev is not None:
                    bucket = bucket.prev
            elif scope > self._bucket.scope:
                next_scope = self._scope_class(self._bucket.scope + 1)
                bucket = self._construct_chain(bucket, next_scope, scope)

        return BucketResolver(self._registry, bucket, **self._filters)

    @override
    def advance(self) -> OperationContext:
        if self._bucket.scope is None:
            return self

        next_val = self._bucket.scope + 1
        if next_val >= len(self._scope_class):
            return self
        next_scope = self._scope_class(next_val)
        return self.context(next_scope)

    def __repr__(self) -> str:
        if self.scope is None:
            res = 'OperationContext'
        else:
            res = f'OperationContext <scope={self.scope.name.lower()}>'

        if self._filters:
            filters_repr = ','.join([f'{k}={v}' for k, v in self._filters.items()])
            res += f', filters={filters_repr}>'

        return res

    def __del__(self) -> None:
        if self._bucket.prev:
            self._bucket.prev.num_children -= 1
            self._bucket.prev = None
        del self._bucket

    @override
    def _create(self) -> Selector[type, Any]:
        return self._data.factory(
            bucket=self._bucket, registry=self._registry, filters=self._filters
        )

    @override
    def _enter(self, instance: _ResolverSelector) -> Selector[type, Any]:
        instance.resolver.__enter__()
        return instance

    @override
    def _exit(
        self,
        instance: _ResolverSelector,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        instance.resolver.__exit__(exc_type, exc_value, traceback)

    def _construct_chain(
        self, bucket: Bucket, scope_from: IntEnum, scope_to: IntEnum
    ) -> Bucket:
        for next_val in range(scope_from, scope_to + 1):
            if next_val >= len(self._scope_class):
                break
            bucket.num_children += 1
            bucket = Bucket(prev=bucket, scope=self._scope_class(next_val))
        return bucket



import enum
from dataclasses import dataclass, field
from datetime import datetime

from greyhorse.enum import Enum, Struct, Unit


class ResourceEventKind(enum.IntFlag):
    SETUP = enum.auto()
    TEARDOWN = enum.auto()
    ALL = SETUP | TEARDOWN


@dataclass(slots=True, frozen=True, kw_only=True)
class Condition:
    name: str
    status: bool = False
    message: str | None = None
    permanent: bool = False
    transition_dt: datetime | None = None
    last_probe_dt: datetime = field(default_factory=datetime.now)
    # metadata: Mapping[str, Any] = field(default_factory=dict)


class TransitionStatus(Enum):
    Pending = Unit()
    Partial = Struct(total=int, completed=int)
    Complete = Unit()
    Failure = Struct(reasons=list[str], message=str)


class ResourceStatus(Enum):
    Idle = Unit()
    Initialization = Unit()
    Finalization = Unit()
    Initialized = Unit()
    Finalized = Unit()
    Failed = Struct(message=str)
    Available = Struct(alive=bool, in_use=bool)
    Acquired = Struct(alive=bool)



from __future__ import annotations

import asyncio
import threading
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable, Mapping
from contextlib import (
    AbstractAsyncContextManager,
    AbstractContextManager,
    AsyncExitStack,
    ExitStack,
    suppress,
)
from dataclasses import dataclass, field
from types import TracebackType
from typing import Any, override
from uuid import uuid4

from greyhorse.app.private.runtime.invoke import (
    invoke_async,
    invoke_sync,
    is_like_async_context_manager,
    is_like_sync_context_manager,
)
from greyhorse.enum import Enum, Struct, Unit
from greyhorse.factory import Factory, FactoryFn, into_factory
from greyhorse.utils.types import TypeWrapper


type FieldFactory[T] = (
    T
    | Callable[[], Awaitable[T] | T]
    | AbstractContextManager[T]
    | AbstractAsyncContextManager[T]
    | Callable[[], AbstractContextManager[T]]
    | Callable[[], AbstractAsyncContextManager[T]]
)


type ContextManagerLike[T] = (
    AbstractContextManager[T]
    | AbstractAsyncContextManager[T]
    | Callable[[], AbstractContextManager[T]]
    | Callable[[], AbstractAsyncContextManager[T]]
)


class ContextState[T](Enum):
    Idle = Unit()
    InUse = Struct(count=int, value=T)
    Applied = Struct(count=int, value=T)
    Cancelled = Struct(count=int, value=T)


@dataclass(slots=True, frozen=True, kw_only=True)
class ContextData[T]:
    factory: Factory[T]
    ident: str = field(default_factory=lambda: str(uuid4()))
    field_factories: dict[str, FieldFactory[Any]] = field(default_factory=dict)
    finalizers: list[Callable[[], Awaitable[None] | None]] = field(default_factory=list)


class InvalidContextStateError(RuntimeError):
    pass


class Context:
    __slots__ = ('_data', '_state')

    def __init__(
        self,
        factory: Factory[Any] | FactoryFn[Any],
        fields: Mapping[str, FieldFactory[Any]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
    ) -> None:
        self._state = ContextState[Any].Idle
        if not isinstance(factory, Factory):
            factory = into_factory(factory)
        self._data = ContextData(
            factory=factory, field_factories=fields or {}, finalizers=finalizers or []
        )

    @property
    def ident(self) -> str:
        return self._data.ident

    @property
    def state(self) -> ContextState:
        return self._state

    def __enter__(self) -> object:
        raise NotImplementedError

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        raise NotImplementedError

    async def __aenter__(self) -> object:
        raise NotImplementedError

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        raise NotImplementedError

    def children(self) -> list[Context]:
        raise NotImplementedError


class MutContext(Context, ABC):
    @abstractmethod
    def apply(self) -> Awaitable[None] | None: ...

    @abstractmethod
    def cancel(self) -> Awaitable[None] | None: ...

    @abstractmethod
    def mut_children(self) -> list[MutContext]: ...


class SyncContext[T](Context, TypeWrapper[T], AbstractContextManager):
    __slots__ = ('_children', '_lock', '_stack', '_sub_contexts')

    def __init__(
        self,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
    ) -> None:
        fields = fields.copy() if fields else {}
        self._stack = ExitStack()
        self._lock = threading.Lock()
        self._children: list[SyncContext] = []
        self._sub_contexts: list[tuple[ContextManagerLike, str | None]] = []

        if sub_contexts:
            for ctx in sub_contexts:
                if isinstance(ctx, SyncContext):
                    self._children.append(ctx)
                if is_like_sync_context_manager(ctx):
                    self._sub_contexts.append((ctx, None))

        names_to_remove = set()

        for name, value in fields.items():
            if isinstance(value, SyncContext):
                self._children.append(value)
                names_to_remove.add(name)
            if is_like_sync_context_manager(value):
                self._sub_contexts.append((value, name))
                names_to_remove.add(name)
            elif is_like_async_context_manager(value):
                names_to_remove.add(name)

        for name in names_to_remove:
            fields.pop(name)

        super().__init__(factory, fields, finalizers)

    @override
    def children(self) -> list[Context]:
        return self._children.copy()

    def _switch_to_use(self) -> T:
        self._stack.__enter__()
        kwargs: dict[str, Any] = {}

        for ctx, field in self._sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if (value := self._stack.enter_context(ctx)) and field is not None:
                kwargs[field] = value

        for name, value in self._data.field_factories.items():
            if callable(value):
                value = invoke_sync(value)
            kwargs[name] = value

        return self._create(**kwargs)

    def _switch_to_idle(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        with suppress(Exception):
            self._destroy(instance)

        for finalizer in self._data.finalizers:
            with suppress(Exception):
                invoke_sync(finalizer)

        # with suppress(Exception):
        self._stack.__exit__(exc_type, exc_value, traceback)

    def _create(self, **kwargs: Any) -> T:
        return invoke_sync(self._data.factory, **kwargs)

    def _destroy(self, instance: T) -> None:
        invoke_sync(self._data.factory.destroy, instance)

    def _enter(self, instance: T) -> T:
        return instance

    def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        pass

    def _nested_enter(self, instance: T) -> T:
        return instance

    def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        pass

    @override
    def __enter__(self) -> T:
        with self._lock:
            match self._state:
                case ContextState.Idle:
                    value = self._switch_to_use()
                    self._state = ContextState[self.wrapped_type].InUse(count=1, value=value)
                    return self._enter(value)

                case (
                    ContextState.InUse(count, value)
                    | ContextState.Applied(count, value)
                    | ContextState.Cancelled(count, value)
                ):
                    self._state = self._state.__class__(count=count + 1, value=value)
                    return self._nested_enter(value)

    @override
    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextStateError('Context exit on idle state')

                case (
                    ContextState.InUse(count, value)
                    | ContextState.Applied(count, value)
                    | ContextState.Cancelled(count, value)
                ):
                    if count > 1:
                        self._nested_exit(value, exc_type, exc_value, traceback)
                        self._state = self._state.__class__(count=count - 1, value=value)
                    else:
                        self._exit(value, exc_type, exc_value, traceback)
                        self._switch_to_idle(value, exc_type, exc_value, traceback)
                        self._state = ContextState[self.wrapped_type].Idle


class SyncMutContext[T](SyncContext[T], MutContext):
    __slots__ = ('_auto_apply', '_force_rollback', '_mut_children')

    def __init__(
        self,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False,
        auto_apply: bool = False,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts)
        self._mut_children = []
        self._force_rollback = force_rollback
        self._auto_apply = auto_apply

        for child in self._children:
            if isinstance(child, SyncMutContext):
                self._mut_children.append(child)

    @override
    def mut_children(self) -> list[MutContext]:
        return self._mut_children.copy()

    def _apply(self, instance: T) -> None:
        pass

    def _cancel(self, instance: T) -> None:
        pass

    @override
    def apply(self) -> None:
        with self._lock:
            self._do_apply()

    @override
    def cancel(self) -> None:
        with self._lock:
            self._do_cancel()

    @override
    def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._force_rollback or exc_type is not None:
            self._do_cancel()
        elif self._auto_apply:
            self._do_apply()

    def _do_apply(self) -> None:
        match self._state:
            case ContextState.Idle:
                raise InvalidContextStateError('MutContext apply on idle state')

            case ContextState.InUse(count, value):
                for child in self._mut_children:
                    child.apply()
                self._apply(value)
                self._state = ContextState[self.wrapped_type].Applied(count=count, value=value)

            case ContextState.Applied(_, _):
                pass

            case ContextState.Cancelled(_, _):
                raise InvalidContextStateError('MutContext apply on cancelled state')

    def _do_cancel(self) -> None:
        match self._state:
            case ContextState.Idle:
                raise InvalidContextStateError('MutContext cancel on idle state')

            case ContextState.InUse(count, value):
                for child in self._mut_children:
                    child.cancel()
                self._cancel(value)
                self._state = ContextState[self.wrapped_type].Cancelled(
                    count=count, value=value
                )

            case ContextState.Applied(_, _):
                raise InvalidContextStateError('MutContext cancel on applied state')

            case ContextState.Cancelled(_, _):
                pass


class AsyncContext[T](Context, TypeWrapper[T], AbstractAsyncContextManager):
    __slots__ = ('_children', '_lock', '_stack', '_sub_contexts')

    def __init__(
        self,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
    ) -> None:
        fields = fields.copy() if fields else {}
        self._stack = AsyncExitStack()
        self._lock = asyncio.Lock()
        self._children: list[Context] = []
        self._sub_contexts: list[tuple[ContextManagerLike[T], bool, str | None]] = []

        if sub_contexts:
            for ctx in sub_contexts:
                if isinstance(ctx, SyncContext):
                    self._children.append(ctx)
                    self._sub_contexts.append((ctx, False, None))
                elif isinstance(ctx, AsyncContext):
                    self._children.append(ctx)
                    self._sub_contexts.append((ctx, True, None))
                elif is_like_sync_context_manager(ctx):
                    self._sub_contexts.append((ctx, False, None))
                elif is_like_async_context_manager(ctx):
                    self._sub_contexts.append((ctx, True, None))

        names_to_remove = set()

        for name, value in fields.items():
            if isinstance(value, SyncContext):
                self._children.append(value)
                self._sub_contexts.append((value, False, name))
                names_to_remove.add(name)
            elif isinstance(value, AsyncContext):
                self._children.append(value)
                self._sub_contexts.append((value, True, name))
                names_to_remove.add(name)
            elif is_like_sync_context_manager(value):
                self._sub_contexts.append((value, False, name))
                names_to_remove.add(name)
            elif is_like_async_context_manager(value):
                self._sub_contexts.append((value, True, name))
                names_to_remove.add(name)

        for name in names_to_remove:
            fields.pop(name)

        super().__init__(factory, fields, finalizers)

    @override
    def children(self) -> list[Context]:
        return self._children.copy()

    async def _switch_to_use(self) -> T:
        await self._stack.__aenter__()
        kwargs: dict[str, Any] = {}

        for ctx, is_async, field in self._sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if is_async:
                value = await self._stack.enter_async_context(ctx)
            else:
                value = self._stack.enter_context(ctx)

            if value is not None and field is not None:
                kwargs[field] = value

        for name, value in self._data.field_factories.items():
            if callable(value):
                value = await invoke_async(value)
            kwargs[name] = value

        return await self._create(**kwargs)

    async def _switch_to_idle(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        with suppress(Exception):
            await self._destroy(instance)

        for finalizer in self._data.finalizers:
            with suppress(Exception):
                await invoke_async(finalizer)

        # with suppress(Exception):
        await self._stack.__aexit__(exc_type, exc_value, traceback)

    async def _create(self, **kwargs: Any) -> T:
        return await invoke_async(self._data.factory, **kwargs)

    async def _destroy(self, instance: T) -> None:
        await invoke_async(self._data.factory.destroy, instance)

    async def _enter(self, instance: T) -> T:
        return instance

    async def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        pass

    async def _nested_enter(self, instance: T) -> T:
        return instance

    async def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        pass

    @override
    async def __aenter__(self) -> T:
        async with self._lock:
            match self._state:
                case ContextState.Idle:
                    value = await self._switch_to_use()
                    self._state = ContextState[self.wrapped_type].InUse(count=1, value=value)
                    return await self._enter(value)

                case (
                    ContextState.InUse(count, value)
                    | ContextState.Applied(count, value)
                    | ContextState.Cancelled(count, value)
                ):
                    self._state = self._state.__class__(count=count + 1, value=value)
                    return await self._nested_enter(value)

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        async with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextStateError('Context exit on idle state')

                case (
                    ContextState.InUse(count, value)
                    | ContextState.Applied(count, value)
                    | ContextState.Cancelled(count, value)
                ):
                    if count > 1:
                        await self._nested_exit(value, exc_type, exc_value, traceback)
                        self._state = self._state.__class__(count=count - 1, value=value)
                    else:
                        await self._exit(value, exc_type, exc_value, traceback)
                        await self._switch_to_idle(value, exc_type, exc_value, traceback)
                        self._state = ContextState[self.wrapped_type].Idle


class AsyncMutContext[T](AsyncContext[T], MutContext):
    __slots__ = ('_auto_apply', '_force_rollback', '_mut_children')

    def __init__(
        self,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False,
        auto_apply: bool = False,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts)
        self._mut_children: list[MutContext] = []
        self._force_rollback = force_rollback
        self._auto_apply = auto_apply

        for child in self._children:
            if isinstance(child, (SyncMutContext, AsyncMutContext)):
                self._mut_children.append(child)

    @override
    def mut_children(self) -> list[MutContext]:
        return self._mut_children.copy()

    async def _apply(self, instance: T) -> None:
        pass

    async def _cancel(self, instance: T) -> None:
        pass

    @override
    async def apply(self) -> None:
        async with self._lock:
            await self._do_apply()

    @override
    async def cancel(self) -> None:
        async with self._lock:
            await self._do_cancel()

    @override
    async def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._force_rollback or exc_type is not None:
            await self._do_cancel()
        elif self._auto_apply:
            await self._do_apply()

    async def _do_apply(self) -> None:
        match self._state:
            case ContextState.Idle:
                raise InvalidContextStateError('MutContext apply on idle state')

            case ContextState.InUse(count, value):
                for child in self._mut_children:
                    await invoke_async(child.apply)

                await self._apply(value)
                self._state = ContextState[self.wrapped_type].Applied(count=count, value=value)

            case ContextState.Applied(_, _):
                pass

            case ContextState.Cancelled(_, _):
                raise InvalidContextStateError('MutContext apply on cancelled state')

    async def _do_cancel(self) -> None:
        match self._state:
            case ContextState.Idle:
                raise InvalidContextStateError('MutContext cancel on idle state')

            case ContextState.InUse(count, value):
                for child in self._mut_children:
                    await invoke_async(child.cancel)

                await self._cancel(value)
                self._state = ContextState[self.wrapped_type].Cancelled(
                    count=count, value=value
                )

            case ContextState.Applied(_, _):
                raise InvalidContextStateError('MutContext cancel on applied state')

            case ContextState.Cancelled(_, _):
                pass


@dataclass(frozen=True, slots=True, kw_only=True)
class CtxCallbacks:
    before_create: Callable[..., Any] | None = None
    after_create: Callable[[Any], Any] | None = None
    before_destroy: Callable[[Any], Any] | None = None
    after_destroy: Callable[[], Any] | None = None
    on_enter: Callable[[Any], Any] | None = None
    on_exit: (
        Callable[
            [Any, type[BaseException] | None, BaseException | None, TracebackType | None], Any
        ]
        | None
    ) = None
    on_nested_enter: Callable[[Any], Any] | None = None
    on_nested_exit: (
        Callable[
            [Any, type[BaseException] | None, BaseException | None, TracebackType | None], Any
        ]
        | None
    ) = None


@dataclass(frozen=True, slots=True, kw_only=True)
class MutCtxCallbacks(CtxCallbacks):
    on_apply: Callable[[Any], Any] | None = None
    on_cancel: Callable[[Any], Any] | None = None


class SyncCallbackContext[T](SyncContext[T]):
    __slots__ = ('_callbacks',)

    def __init__(
        self,
        callbacks: CtxCallbacks,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts)
        self._callbacks = callbacks

    @override
    def _create(self, **kwargs: Any) -> T:
        if self._callbacks.before_create is not None:
            self._callbacks.before_create(**kwargs)
        instance = super()._create(**kwargs)
        if self._callbacks.after_create is not None:
            self._callbacks.after_create(instance)
        return instance

    @override
    def _destroy(self, instance: T) -> None:
        if self._callbacks.before_destroy is not None:
            self._callbacks.before_destroy(instance)
        super()._destroy(instance)
        if self._callbacks.after_destroy is not None:
            self._callbacks.after_destroy()

    @override
    def _enter(self, instance: T) -> T:
        if self._callbacks.on_enter is not None:
            return self._callbacks.on_enter(instance)
        return instance

    @override
    def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._callbacks.on_exit is not None:
            self._callbacks.on_exit(instance, exc_type, exc_value, traceback)

    @override
    def _nested_enter(self, instance: T) -> T:
        if self._callbacks.on_nested_enter is not None:
            return self._callbacks.on_nested_enter(instance)
        return instance

    @override
    def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._callbacks.on_nested_exit is not None:
            self._callbacks.on_nested_exit(instance, exc_type, exc_value, traceback)


class SyncMutCallbackContext[T](SyncMutContext[T]):
    __slots__ = ('_callbacks',)

    def __init__(
        self,
        callbacks: MutCtxCallbacks,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False,
        auto_apply: bool = False,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts, force_rollback, auto_apply)
        self._callbacks = callbacks

    @override
    def _create(self, **kwargs: Any) -> T:
        if self._callbacks.before_create is not None:
            self._callbacks.before_create(**kwargs)
        instance = super()._create(**kwargs)
        if self._callbacks.after_create is not None:
            self._callbacks.after_create(instance)
        return instance

    @override
    def _destroy(self, instance: T) -> None:
        if self._callbacks.before_destroy is not None:
            self._callbacks.before_destroy(instance)
        super()._destroy(instance)
        if self._callbacks.after_destroy is not None:
            self._callbacks.after_destroy()

    @override
    def _enter(self, instance: T) -> T:
        if self._callbacks.on_enter is not None:
            return self._callbacks.on_enter(instance)
        return instance

    @override
    def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._callbacks.on_exit is not None:
            self._callbacks.on_exit(instance, exc_type, exc_value, traceback)

    @override
    def _nested_enter(self, instance: T) -> T:
        if self._callbacks.on_nested_enter is not None:
            self._callbacks.on_nested_enter(instance)
        return instance

    @override
    def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._callbacks.on_nested_exit is not None:
            self._callbacks.on_nested_exit(instance, exc_type, exc_value, traceback)

    @override
    def _apply(self, instance: T) -> None:
        if self._callbacks.on_apply is not None:
            self._callbacks.on_apply(instance)

    @override
    def _cancel(self, instance: T) -> None:
        if self._callbacks.on_cancel is not None:
            self._callbacks.on_cancel(instance)


class AsyncCallbackContext[T](AsyncContext[T]):
    __slots__ = ('_callbacks',)

    def __init__(
        self,
        callbacks: CtxCallbacks,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts)
        self._callbacks = callbacks

    @override
    async def _create(self, **kwargs: Any) -> T:
        if self._callbacks.before_create is not None:
            self._callbacks.before_create(**kwargs)
        instance = await super()._create(**kwargs)
        if self._callbacks.after_create is not None:
            self._callbacks.after_create(instance)
        return instance

    @override
    async def _destroy(self, instance: T) -> None:
        if self._callbacks.before_destroy is not None:
            self._callbacks.before_destroy(instance)
        await super()._destroy(instance)
        if self._callbacks.after_destroy is not None:
            self._callbacks.after_destroy()

    @override
    async def _enter(self, instance: T) -> T:
        if self._callbacks.on_enter is not None:
            return self._callbacks.on_enter(instance)
        return instance

    @override
    async def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._callbacks.on_exit is not None:
            self._callbacks.on_exit(instance, exc_type, exc_value, traceback)

    @override
    async def _nested_enter(self, instance: T) -> T:
        if self._callbacks.on_nested_enter is not None:
            return self._callbacks.on_nested_enter(instance)
        return instance

    @override
    async def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._callbacks.on_nested_exit is not None:
            self._callbacks.on_nested_exit(instance, exc_type, exc_value, traceback)


class AsyncMutCallbackContext[T](AsyncMutContext[T]):
    __slots__ = ('_callbacks',)

    def __init__(
        self,
        callbacks: MutCtxCallbacks,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False,
        auto_apply: bool = False,
    ) -> None:
        super().__init__(factory, fields, finalizers, sub_contexts, force_rollback, auto_apply)
        self._callbacks = callbacks

    @override
    async def _create(self, **kwargs: Any) -> T:
        if self._callbacks.before_create is not None:
            self._callbacks.before_create(**kwargs)
        instance = await super()._create(**kwargs)
        if self._callbacks.after_create is not None:
            self._callbacks.after_create(instance)
        return instance

    @override
    async def _destroy(self, instance: T) -> None:
        if self._callbacks.before_destroy is not None:
            self._callbacks.before_destroy(instance)
        await super()._destroy(instance)
        if self._callbacks.after_destroy is not None:
            self._callbacks.after_destroy()

    @override
    async def _enter(self, instance: T) -> T:
        if self._callbacks.on_enter is not None:
            return self._callbacks.on_enter(instance)
        return instance

    @override
    async def _exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._callbacks.on_exit is not None:
            self._callbacks.on_exit(instance, exc_type, exc_value, traceback)

    @override
    async def _nested_enter(self, instance: T) -> T:
        if self._callbacks.on_nested_enter is not None:
            self._callbacks.on_nested_enter(instance)
        return instance

    @override
    async def _nested_exit(
        self,
        instance: T,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        if self._callbacks.on_nested_exit is not None:
            self._callbacks.on_nested_exit(instance, exc_type, exc_value, traceback)

    @override
    async def _apply(self, instance: T) -> None:
        if self._callbacks.on_apply is not None:
            self._callbacks.on_apply(instance)

    @override
    async def _cancel(self, instance: T) -> None:
        if self._callbacks.on_cancel is not None:
            self._callbacks.on_cancel(instance)


class ContextBuilder[T](TypeWrapper[T]):
    def __init__[**P](
        self,
        factory: Factory[T] | FactoryFn[T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        sub_contexts: list[ContextManagerLike[T]] | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        self._factory = factory
        self._fields = fields or {}
        self._finalizers = finalizers or []
        self._sub_contexts = sub_contexts or []
        self._kwargs = kwargs

    def add_param(self, name: str, value: FieldFactory[T]) -> None:
        self._fields[name] = value

    def add_sub_context(self, context: ContextManagerLike[T]) -> None:
        self._sub_contexts.append(context)

    def add_finalizer(self, finalizer: Callable[[], Awaitable[None] | None]) -> None:
        self._finalizers.append(finalizer)

    def build(self) -> T:
        return self.wrapped_type(
            factory=self._factory,
            fields=self._fields,
            finalizers=self._finalizers,
            sub_contexts=self._sub_contexts,
            **self._kwargs,
        )

    def __class_getitem__[C: SyncContext | AsyncContext | SyncMutContext | AsyncMutContext](
        cls, args: tuple[type[C], type[T]]
    ) -> ContextBuilder:
        class_, type_ = args
        return super().__class_getitem__(class_[type_])



from abc import ABC, abstractmethod
from typing import Any, Self


class Query(ABC):
    __slots__ = ('_filter_exprs', '_filter_params', '_sorting_exprs')

    def __init__(
        self,
        filter_exprs: list[Any] | None = None,
        sorting_exprs: list[Any] | None = None,
        **filter_params: dict[str, Any],
    ) -> None:
        self._filter_exprs = filter_exprs or []
        self._filter_params = filter_params
        self._sorting_exprs = sorting_exprs or []

    def add_filter[F](self, filter: F) -> 'Query':
        filter_exprs = [*self._filter_exprs, filter]
        return type(self)(filter_exprs, self._sorting_exprs, **self._filter_params)

    def add_sorting[S](self, sorting: S) -> 'Query':
        sorting_exprs = [*self._sorting_exprs, sorting]
        return type(self)(self._filter_exprs, sorting_exprs, **self._filter_params)

    def clone(self) -> 'Query':
        return type(self)(self._filter_exprs[:], self._sorting_exprs[:], **self._filter_params)

    @abstractmethod
    def apply_filter[F](self, filter: F) -> F: ...

    @abstractmethod
    def apply_sorting[S](self, sorting: S) -> S: ...

    def __eq__(self, other: Self) -> bool:
        return (self._filter_exprs, self._filter_params, self._sorting_exprs) == (
            other._filter_exprs,
            other._filter_params,
            other._sorting_exprs,
        )

    def __ne__(self, other: Self) -> bool:
        return not (self == other)
